from astropy.io import fits
from astropy.nddata import NDData
from astropy.stats import sigma_clipped_stats
from astropy.table import Table
import matplotlib.pyplot as plt
import numpy as np
import os
from photutils.detection import DAOStarFinder, IRAFStarFinder
from photutils.psf import extract_stars, EPSFBuilder, GriddedPSFModel
import pkg_resources
from spike import tools
import subprocess
import urllib
import warnings

try: #for version control reasons
	import stpsf as webbpsf
except:
	import webbpsf

def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
	return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)

warnings.formatwarning = warning_on_one_line

CONFIG_PATH = pkg_resources.resource_filename('spike', 'configs/')

tinyparams = {}
tinyparams['imcam'] = {'WFPC1/WFC':1, 'WFPC1/PC':2, 'FOC/f48':3, 'FOC/f96':4, 
'WFPC2/WFC':5, 'WFPC2/PC':6, 'FOC/f48-COSTAR':7, 'FOC/f96-COSTAR':8, 'NICMOS/NIC1-precool':9,
'NICMOS/NIC2-precool':10, 'NICMOS/NIC3-precool':11, 'STIS/CCD':12, 'STIS/NUV':13, 'STIS/FUV':14,
'ACS/WFC':15, 'ACS/HRC':16, 'ACS/HRC-offspot':17, 'ACS/SBC':18, 'NICMOS/NIC1':19, 
'NICMOS/NIC2':20, 'NICMOS/NIC3':21, 'WFC3/UVIS':22, 'WFC3/IR':23}
# listing all options here, though not all are explicitly included in the spike code
tinyparams['specparam'] = {'O5':1, 'O8F':2, 'O6':3, 'B1V':4, 'B3V':5, 'B6V':6, 'A05':7, 'A5V':8, 
'F6V':9, 'F8V':10, 'G2V':11, 'G5V':12, 'G8V':13, 'K4V':14, 'K7V':15, 'M1.5V':16, 'M3V':17}

try:
	TINY_PATH = os.environ['TINYTIM']
except:
	TINY_PATH = None


stdpsf_jwdet = {'NRCA1':1, 'NRCA2':2, 'NRCA3':3, 'NRCA4':4, 
'NRCB1':5, 'NRCB2':6, 'NRCB3':7, 'NRCB7':8} #for STDPSFs

plate_scale = {'ACS/WFC':0.05, 'ACS/HRC':0.025 , 'WFC3/IR':0.13, 'WFC3/UVIS':0.039, 'WFPC_wf':0.1016 , 
'WFPC_pc':0.0439, 'WFPC1_wf':0.1016, 'WFPC1_pc':0.0439, 'WFPC2_wf':0.1, 'WFPC2_pc':0.046, 
'NIRCAM_long':0.063, 'NIRCAM_short':0.031, 'MIRI':0.11, 'NIRISS/Imaging':0.066, 'NIRISS':0.066,
'WFI':0.11, 'CGI':0.0218} #arseconds/pixel -- some are averaged across filter/position
# including both WFPC1 and WFPC so that either name can be used

##########
# * * * *
##########


def tinypsf(coords, img, imcam, pos, plot = False, verbose = False, writeto = True,
	ebmv = None, av = None, wmag = None,
	jitter = None, major = None, minor = None, angle = None,
	specchoice = 'blackbody', listchoice = 'G5V', temp = 6000., 
	specalpha = 1., specbeta = 1., fov_arcsec = 6., despace = 0.):
	"""
	Generate HST PSFs using TinyTim.
	All of the options from TinyTim are easily available here *except* custom filters
	and subsampling, as they complicate use with spike. If you would like a subsampled
	Tiny Tim PSF model, please use spike.psfgen.tinygillispsf() instead.

	Parameters:
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		img (str): Path to image for which PSF is generated.
		imcam (str): Specification of instrument/camera used to capture the images (e.g., 'ACS/WFC', 'WFC3/IR', 'WFPC', 'WFPC2').
			For 'WFPC' and 'WFPC2', the camera is selected by-chip and should not be specified here.
		pos (list): Location of object of interest (spatial and spectral). [X, Y, chip, filter]
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		writeto (bool): If True, will write 2D model PSF (differentiated with '_topsf' 
			suffix) and will amend relevant image WCS information/remove extraneous extensions.
			This is in addition to the 2D PSF model saved by TinyTim.
		ebmv (float): Interstellar extinction, specified by reddening in mag. 
			If specified, av should not be.
		av (float): Interstellar extinction, specified by visual extinction in mag. 
			If specified, ebmv should not be.
		wmag (float): Fraction of wavelengths to use in generating polychromatic PSFs 
			(if wmag > 1, uses more than the default; if wmag < 1, uses fewer).
		jitter (float): Symmetric jitter in mas.
		major (float): Axisymmetric major axis jitter in mas.
		minor (float): Axisymmetric minor axis jitter in mas.
		angle (float): Angle of offset for axisymmetric jitter.
		specchoice (str): 'list', 'blackbody', 'plaw_nu', 'plaw_lam' -- if 'list', must also specify
			listchoice; if 'blackbody', must also specify temp; if 'plaw_nu', must also specify specalpha;
			and if 'plaw_lam', must also specify specbeta.
		listchoice (str): One of 'O5', 'O8F', 'O6', 'B1V', 'B3V', 'B6V', 'A0V', 'A5V', 'F6V', 'F8V',
			'G2V', 'G5V', 'G8V', 'K4V', 'K7V', 'M1.5V', 'M3V'
		temp (float): Temperature of blackbody spectrum in K.
		specalpha (float): Spectral index alpha for F(nu)~nu^alpha.
		specbeta (float): Spectral index alpha for F(lambda)~lambda^beta.
		fov_arcsec (float): "Diameter" of model PSF image in arcsec.
		despace (float): Focus, secondary mirror despace in micron. Scaled by 0.011 and added to
			the 4th Zernike polynomial.

	Returns:
		TinyTim model PSF
	"""
	if not TINY_PATH:
		# this is a warning and not an error on the off chance that the TinyTim executables are 
		# in the working directory
		warnings.warn('Tiny Tim is not in your path. Make sure that it is installed -- https://github.com/spacetelescope/tinytim/releases/tag/7.5 -- and your TINYTIM environment variable is set or select a different PSF generation mode.', Warning, stacklevel = 2)

	tiny1 = TINY_PATH+'/tiny1 tiny.param'

	if ebmv and av:
		warnings.warn("Only one of ebmv and Av can be specified. Proceeding using only ebmv.", Warning, stacklevel = 2)
		tiny1 += ' ebmv='+str(ebmv)
	
	if None in [ebmv, av]:
		if ebmv:
			tiny1 += ' ebmv='+str(ebmv)

		if av:
			tiny1 += ' av='+str(av)

	if wmag:
		tiny1 += ' wmag'+str(wmag)

	if jitter:
		tiny1 += ' jitter='+str(jitter)

	if major and minor and angle:
		tiny1 += ' major=%s minor=%s angle=%s'%(str(major), str(minor), str(angle))

	if None in [major, minor, angle]:
		warnings.warn('All of major, minor, and angle must be specified to be applied. Proceeding with no elliptical jitter.', Warning, stacklevel = 2)

	if specchoice == 'list':
		spec = 1
		specparam = tinyparams['specparam'][listchoice]

	if specchoice == 'blackbody':
		spec = 2
		specparam = temp

	if specchoice == 'plaw_fnu':
		spec = 3
		specparam = specalpha

	if specchoice == 'plaw_flam':
		spec = 4
		specparam = specbeta

	coordstring = str(coords.ra)
	if coords.dec.deg >= 0:
		coordstring += '+'+str(coords.dec)
	if coords.dec.deg < 0:
		coordstring += str(coords.dec)

	modname = img.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_psf')

	if imcam in ['ACS/WFC', 'WFC3/UVIS']:
		command_list = [tinyparams['imcam'][imcam], pos[2], '%i %i'%(pos[0], pos[1]), 
		pos[3], spec, specparam, fov_arcsec, despace, modname]
	if imcam in ['ACS/HRC', 'WFC3/IR']:
		command_list = [16, '%i %i'%(pos[0], pos[1]), pos[3], 
		spec, specparam, fov_arcsec, despace, modname]
	if imcam in ['WFPC1', 'WFPC']:
		if pos[2] <= 4:
			imcam = 'WFPC/WFC'
		if pos[2] >= 5:
			imcam = 'WFPC/PC'
		imfits = fits.open(img)
		yyyy, mm, dd = imfits[0].header['DATE'].split('T')[0].split('-')
		command_list = [tinyparams['imcam'][imcam], pos[2], '%i %i'%(pos[0], pos[1]), 
		'%i %i %i'%(dd, mm, yyyy), pos[3], spec, specparam, fov_arcsec, 'N', despace, modname]
	if (imcam == 'WFPC2') and (pos[2] == 1):
		imcam = 'WFPC2/PC'
		command_list = [tinyparams['imcam'][imcam], '%i %i'%(pos[0], pos[1]), 
		pos[3], spec, specparam, fov_arcsec, 'N', despace, modname]
	if (imcam == 'WFPC2') and (pos[2] >= 2):
		imcam = 'WFPC2/WFC'
		command_list = [tinyparams['imcam'][imcam], pos[2], '%i %i'%(pos[0], pos[1]), 
		pos[3], spec, specparam, fov_arcsec, 'N', despace, modname]


	if verbose:
		print('Generating parameter file.')

	commandlist = [str(clt) for clt in command_list]

	tiny = subprocess.Popen(tiny1, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
	newline = os.linesep

	tiny.communicate(newline.join(commandlist).encode())

	
	os.system(TINY_PATH+'/tiny2 tiny.param')
	if verbose:
		print("Completed PSF modeling.")

	if imcam in ['ACS/WFC', 'ACS/HRC']: 
		os.system(TINY_PATH+'/tiny3 tiny.param')
		if verbose:
			print("Completed geometric distortion correction.")

	# deal with NICMOS and STIS -- will add later
	# inclined to include NICMOS but not STIS
	# NICMOS has time complication since the cryo-cooler etc. matter
	# set date of change over at April 3, 2002: https://ui.adsabs.harvard.edu/abs/2008AIPC..985..799S/abstract

	if imcam not in  ['ACS/WFC', 'ACS/HRC']:
		psfmodel = fits.open(modname+'00_psf.fits')[0].data

	if imcam in ['ACS/WFC', 'ACS/HRC']:
		psfmodel = fits.open(modname+'00.fits')[0].data


	if plot:
		fig= plt.figure(figsize = (5, 5))
		plt.imshow(psfmodel, origin = 'lower', cmap = 'Greys', 
			vmin = np.nanpercentile(psfmodel, 20), vmax = np.nanpercentile(psfmodel, 97))
		plt.colorbar()
		fig.savefig(modname+'.png', bbox_inches = 'tight', dpi = 100)

		if verbose:
			print('PSF model image written to %s.png'%(modname))

	if writeto:
		if verbose:
			print('Writing to %s.fits.'%modname.replace('_psf', '_topsf'))
		tools.rewrite_fits(psfmodel, coords, img, imcam, pos, method = 'TinyTim')


	return psfmodel



def tinygillispsf(coords, img, imcam, pos, plot = False, verbose = False, keep = False, 
	writeto = True, specchoice = 'blackbody', listchoice = 'G5V', temp = 6000., specalpha = 1., 
	specbeta = 1., fov_arcsec = 6., despace = 0., sample = 1., linearfit = False, regrid = True):
	"""
	Generate HST PSFs using TinyTim and the parameter changes laid out in Gillis et al. (2020), 
	which were tested on ACS imaging. As such, this method is not recommended for instruments
	other than ACS or cameras other than WFC.
	

	Note that the Gillis et al. (2020) code will be downloaded to your working directory. If keep = False (default), 
	it will be removed after use.

	Parameters:
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		img (str): Path to image for which PSF is generated.
		imcam (str): 'ACS/WFC' is the only recommended instrument/camera combination for this PSF generation method.
		pos (list): Location of object of interest (spatial and spectral).[X, Y, chip, filter]
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		keep (bool): If True, retains make_psf.py (Gillis et al. 2020)
		writeto (bool): If True, will write 2D model PSF (differentiated with '_topsf' 
			suffix) and will amend relevant image WCS information/remove extraneous extensions.
			This is in addition to the 2D PSF model saved by TinyTim.
		specchoice (str): 'list', 'blackbody', 'plaw_nu', 'plaw_lam' -- if 'list', must also specify
			listchoice; if 'blackbody', must also specify temp; if 'plaw_nu', must also specify specalpha;
			and if 'plaw_lam', must also specify specbeta.
		listchoice (str): One of 'O5', 'O8F', 'O6', 'B1V', 'B3V', 'B6V', 'A0V', 'A5V', 'F6V', 'F8V',
			'G2V', 'G5V', 'G8V', 'K4V', 'K7V', 'M1.5V', 'M3V'
		temp (float): Temperature of blackbody spectrum in K.
		specalpha (float): Spectral index alpha for F(nu)~nu^alpha.
		specbeta (float): Spectral index alpha for F(lambda)~lambda^beta.
		fov_arcsec (float): "Diameter" of model PSF image in arcsec.
		despace (float): Focus, secondary mirror despace in micron. Scaled by 0.011 and added to
			the 4th Zernike polynomial.
		sample (float): Factor by which to undersample the PSF. Default is not to undersample.
		linearfit (bool): Use linear fit rather than Gillis et al. amended Zernike polynomials.
		regrid (bool): If True, will (interpolate and) regrid model PSF to image pixel scale.

	Returns:
		TinyTim model PSF using amended Gillis et al. (2020) parameters
	"""

	if not TINY_PATH:
		# this is a warning and not an error on the off chance that the TinyTim executables are 
		# in the working directory
		warnings.warn('Tiny Tim is not in your path. Make sure that it is installed -- https://github.com/spacetelescope/tinytim/releases/tag/7.5 -- and your TINYTIM environment variable is set or select a different PSF generation mode.', Warning, stacklevel = 2)

	if not os.path.exists('make_psf.py'):
		# download Gillis et al. (2020) code; https://bitbucket.org/brgillis/tinytim_psfs/src/master/
		rawurl = 'https://bitbucket.org/brgillis/tinytim_psfs/raw/55299ae1a9b3c299b7910a14622e88c0ffd9d8a1/make_psf.py'
		urllib.request.urlretrieve(rawurl, 'make_psf.py')
		if verbose:
			print('Retrieved make_psf.py')


	if specchoice == 'list':
		spec = 1
		specparam = tinyparams['specparam'][listchoice]

	if specchoice == 'blackbody':
		spec = 2
		specparam = temp

	if specchoice == 'plaw_fnu':
		spec = 3
		specparam = specalpha

	if specchoice == 'plaw_flam':
		spec = 4
		specparam = specbeta

	speccall = (spec, specparam)
	
	from make_psf import make_subsampled_model_psf

	coordstring = str(coords.ra)
	if coords.dec.deg >= 0:
		coordstring += '+'+str(coords.dec)
	if coords.dec.deg < 0:
		coordstring += str(coords.dec)
		
	modname = img.replace('.fits', coordstring+'_%s'%pos[3]+'_psf')

	if verbose:
		print('Generating model PSF')

	make_subsampled_model_psf(modname+'.fits',
		    psf_position=(pos[0], pos[1]),
		    focus=despace,
		    chip=pos[2],
		    spec_type=speccall,
		    detector=tinyparams['imcam'][imcam],
		    filter_name=pos[3],
		    psf_size=fov_arcsec,
		    tinytim_path=TINY_PATH,
		    subsampling_factor=sample,
		    linear_fit=linearfit,
		    clobber=True)


	if os.path.exists('make_psf.py') and not keep:
		os.remove('make_psf.py')
		if verbose:
			print('Removed make_psf.py')

	try:
		psfmodel = fits.open(modname+'.fits')[0].data
	except:
		if imcam not in  ['ACS/WFC', 'ACS/HRC']:
			psfmodel = fits.open(modname+'00_psf.fits')[0].data

		if imcam in ['ACS/WFC', 'ACS/HRC']:
			psfmodel = fits.open(modname+'00.fits')[0].data


	if regrid:
		if sample != 1.:
			if verbose:
				print('Regridding PSF model to instrument resolution')
			psfmodel = tools.regridarr(psfmodel, sample)

	if plot:
		fig= plt.figure(figsize = (5, 5))
		plt.imshow(psfmodel, origin = 'lower', cmap = 'Greys', 
			vmin = np.nanpercentile(psfmodel, 20), vmax = np.nanpercentile(psfmodel, 97))
		plt.colorbar()
		fig.savefig(modname+'.png', bbox_inches = 'tight', dpi = 100)

		if verbose:
			print('PSF model image written to %s.png'%(modname))

	if writeto:
		if verbose:
			print('Writing to %s.fits.'%modname.replace('_psf', '_topsf'))
		tools.rewrite_fits(psfmodel, coords, img, imcam, pos, method = 'TinyTim (Gillis+ mod)')

	return psfmodel


def stdpsf(coords, img, imcam, pos, plot = False, verbose = False, 
	writeto = True, fov_arcsec = 6, norm = 1, regrid = True):
	"""
	Coordinate-specific PSFs from STDPSF model grids for HST, JWST.

	Makes use of https://www.stsci.edu/~jayander/HST1PASS/LIB/PSFs/STDPSFs/ and 
	https://www.stsci.edu/~jayander/JWST1PASS/LIB/PSFs/STDPSFs/.

	Parameters:
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		img (str): Path to image for which PSF is generated.
		imcam (str): Specification of instrument/camera used to capture the images (e.g., 'ACS/WFC', 'WFC3/IR', 'WFPC', 
			'WFPC2', 'MIRI', 'NIRCAM', 'NIRISS/Imaging'). For 'WFPC' and 'WFPC2', the camera is selected by-chip and 
			should not be specified here. If 'NIRISS' specified alone, assumes the imaging mode.
		pos (list): Location of object of interest (spatial and spectral).[X, Y, chip, filter]
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		writeto (bool): If True, will write 2D model PSF (differentiated with '_topsf' 
			suffix) and will amend relevant image WCS information/remove extraneous extensions.
		fov_arcsec (float): "Diameter" of model PSF image in arcsec.
		norm (float): Flux normalization for output PSF model.

	Returns:
		STDPSF model PSF

	"""

	# build the url that points to the STDPSF
	imcamurl = imcam.replace('/', '')
	if imcamurl == 'WFC3UVIS':
		imcamurl = 'WFC3UV'

	if imcamurl == 'NIRCAM':
		imcamurl = 'NIRCam'

	if imcamurl == 'NIRISSImaging':
		imcamurl = 'NIRISS'


	if imcam in ['WFPC', 'WFPC1']:
		raise ValueError("There is no available STDPSF grid for WFPC imaging. Please select a different PSF generation method.")

	if imcam in ['ACS/WFC', 'ACS/HRC', 'WFC3/UVIS', 'WFC3/IR', 'WFPC2']:
		baseurl = 'https://www.stsci.edu/~jayander/HST1PASS/LIB/PSFs/STDPSFs/'

		url = baseurl+imcamurl+'/STDPSF_%s_%s'%(imcamurl, pos[3])
		
		if imcam == 'ACS/WFC':
			url += '_SM3.fits'
		elif (imcam == 'WFC3/IR') and (pos[3] == 'F139M'):
			url += '_1x1.fits'
		else:
			url += '.fits'


	if imcam in ['NIRCAM', 'MIRI', 'NIRISS', 'NIRISS/Imaging']:
		baseurl = 'https://www.stsci.edu/~jayander/JWST1PASS/LIB/PSFs/STDPSFs/'

		if imcam in ['MIRI', 'NIRISS', 'NIRISS/Imaging']:
			url = baseurl+imcamurl+'/STDPSF_%s_%s.fits'%(imcamurl, pos[3])

		if imcam == 'NIRCAM':
			if pos[3] in ['F250M', 'F277W', 'F356W', 'F360M', 'F410M', 'F444W', 'F480M']:
				det = pos[2][:-1]+'L'
				url = baseurl+imcamurl+'/LWC/STDPSF_%s_%s.fits'%(det, pos[3])

			if pos[3] in ['F070W', 'F090W', 'F115W', 'F140M', 'F150W', 'F182M', 'F200W'
							'F210M', 'F212N']:
				url = baseurl+imcamurl+'/SWC/%s/STDPSF_%s_%s.fits'%(pos[3], pos[2], pos[3])

	det = None #set detector for photutils
	if imcam in ['WFPC2', 'ACS/WFC', 'WFC3/UVIS']:
		det = pos[2]
	if (imcam == 'NIRCAM') and (pos[2] not in ['NGCA5', 'NRCB5']):
		det = stdpsf_jwdet[pos[2]]

	coordstring = str(coords.ra)
	if coords.dec.deg >= 0:
		coordstring += '+'+str(coords.dec)
	if coords.dec.deg < 0:
		coordstring += str(coords.dec)

	modname = img.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_psf')


	pixkey = imcam #set pixel scale to get the dimensions of x, y
	if imcam in ['WFPC', 'WFPC1']:
		if pos[2] <= 4:
			pixkey += '_wf'
		if pos[2] >= 5:
			pixkey += '_pc'
	if (imcam == 'WFPC2') and (pos[2] == 1):
		pixkey += '_pc'
	if (imcam == 'WFPC2') and (pos[2] >= 2):
		pixkey += 'wf'
	if imcam == 'NIRCAM':
		if pos[2] in ['NRCA5', 'NRCB5', 'NRCALONG', 'NRCBLONG']:
			pixkey += '_long'
		if pos[2] not in ['NRCA5', 'NRCB5', 'NRCALONG', 'NRCBLONG']:
			pixkey += '_short'

	dimxy = fov_arcsec/plate_scale[pixkey] #make square PSF
	halfdim = dimxy//2
	xmin = int(pos[0]) - halfdim
	xmax = int(pos[0]) + halfdim
	ymin = int(pos[1]) - halfdim
	ymax = int(pos[1]) + halfdim

	x, y = np.meshgrid(np.arange(xmin, xmax+1), np.arange(ymin, ymax+1))

	#preferred equivalent to using photutils.psf.stdpsf_reader directly

	model = GriddedPSFModel.read(filename = url, detector_id = det, format= 'stdpsf')
	
	if verbose:
		print('Finished reading STDPSF grid, generating PSF model')
	psfmodel = model.evaluate(x = x, y = y, flux = norm, x_0 = int(pos[0]), y_0 = int(pos[1]))

	if plot:
		fig= plt.figure(figsize = (5, 5))
		plt.imshow(psfmodel, origin = 'lower', cmap = 'Greys', 
			vmin = np.nanpercentile(psfmodel, 20), vmax = np.nanpercentile(psfmodel, 97))
		plt.colorbar()
		fig.savefig(modname+'.png', bbox_inches = 'tight', dpi = 100)

		if verbose:
			print('PSF model image written to %s.png'%(modname))

	if writeto:
		if verbose:
			print('Writing to %s.fits.'%modname.replace('_psf', '_topsf'))
		tools.rewrite_fits(psfmodel, coords, img, imcam, pos, method = 'STDPSFs')


	return psfmodel


def jwpsf(coords, img, imcam, pos, plot = False, verbose = False, writeto = True,
	fov_arcsec = 6, sample = 1., regrid = True, image_mask = None, pupil_mask = None,
	**calckwargs):
	"""
	Generate JWST and Roman PSFs using WebbPSF/STPSF. Note: the WebbPSF name is maintained 
	here in lieu of STPSF to avoid confusion with the generation of empirical STDPSFs.

	Parameters:
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		img (str): Path to image for which PSF is generated.
		imcam (str): Specification of instrument/camera used to capture the images (e.g., 'ACS/WFC', 'WFC3/IR', 'WFPC', 
			'WFPC2', 'MIRI', 'NIRCAM', 'NIRISS/Imaging'). For 'WFPC' and 'WFPC2', the camera is selected by-chip and 
			should not be specified here. If 'NIRISS' specified alone, assumes the imaging mode.
		pos (list): Location of object of interest (spatial and spectral).[X, Y, chip, filter]
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		writeto (bool): If True, will write 2D model PSF (differentiated with '_topsf' 
			suffix) and will amend relevant image WCS information/remove extraneous extensions.
			This is in addition to the 2D PSF models saved by WebbPSF (which will be saved as img_psf.fits).
		fov_arcsec (float): "Diameter" of model PSF image in arcsec.
		sample (float): Factor by which to oversample the PSF.
		regrid (bool): If True, will (interpolate and) regrid model PSF to image pixel scale.
		image_mask (str): Image mask argument for WebbPSF.
		pupil_mask (str): Pupil mask argument for WebbPSF.
		**calckwargs: Additional arguments for calc_psf() -- see 
			https://stpsf.readthedocs.io/en/latest/usage.html#.
			Should be fed to spike.psf.jwst/roman in kwargs as a dictionary called calckwargs.

	Returns:
		WebbPSF model PSF

	"""
	x, y, chip, filt = pos

	coordstring = str(coords.ra)
	if coords.dec.deg >= 0:
		coordstring += '+'+str(coords.dec)
	if coords.dec.deg < 0:
		coordstring += str(coords.dec)

	modname = img.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_psf')

	if imcam.upper() == 'NIRCAM':
		psf = webbpsf.NIRCam()

	if imcam.upper() == 'MIRI':
		psf = webbpsf.MIRI()

	if imcam.upper() == 'NIRISS':
		psf = webbpsf.NIRISS()

	if imcam.upper() == 'WFI':
		psf = webbpsf.roman.WFI()

	if imcam.upper() == 'CGI':
		psf = webbpsf.roman.RomanCoronagraph()
		warnings.warn("WebbPSF Roman CGI development halted in 2017. Use with caution.", Warning, stacklevel = 2) 

	if imcam.upper() in ['NIRCAM', 'WFI']:
		psf.detector = chip


	psf.filter = filt
	psf.detector_position = (x, y)

	if image_mask:
		psf.image_mask = image_mask

	if pupil_mask:
		psf.pupil_mask = pupil_mask

	#read aperture name from FITS header
	imfits = fits.open(img)
	aperturename = imfits[0].header['APERNAME']
	psf.aperturename = aperturename

	if verbose:
		print('Producing PSF model')
	psfmod = psf.calc_psf(fov_arcsec = fov_arcsec, oversample = sample, **calckwargs)

	psfmodel = psfmod[3].data

	if regrid:
		if sample != 1.:
			if verbose:
				print('Regridding PSF to instrument resolution')
			psfmodel = tools.regridarr(psfmodel, sample)


	if plot:
		fig= plt.figure(figsize = (5, 5))
		plt.imshow(psfmodel, origin = 'lower', cmap = 'Greys', 
			vmin = np.nanpercentile(psfmodel, 20), vmax = np.nanpercentile(psfmodel, 97))
		plt.colorbar()
		fig.savefig(modname+'.png', bbox_inches = 'tight', dpi = 100)

		if verbose:
			print('PSF model image written to %s.png'%(modname))


	if writeto:
		if verbose:
			print('Writing to %s.fits.'%modname.replace('_psf', '_topsf'))
		tools.rewrite_fits(psfmodel, coords, img, imcam, pos, method = 'WebbPSF')


	return psfmodel

def effpsf(coords, img, imcam, pos, plot = False, verbose = False, mask = True, writeto = True, 
	fov_arcsec = 6, norm = 1., starselect = 'DAO', starselectargs = {'fwhm':10}, thresh = 125,
	usermask = None, maskval = None, epsfargs = {'oversampling':1, 'progress_bar':True, 'maxiters':10}, ):
	"""
	Generate PSFs using the empirical photutils.epsf routine. 

	NOTE: For frames with fewer stars, this method is prone to artifacts. This can be 
	ameliorated by using a lower detection threshold, but the overall quality of the output 
	PSF may be lower.

	Parameters:
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		img (str): Path to image for which PSF is generated.
		imcam (str): Specification of instrument/camera used to capture the images (e.g., 'ACS/WFC', 'WFC3/IR', 'WFPC', 
			'WFPC2', 'MIRI', 'NIRCAM', 'NIRISS/Imaging'). For 'WFPC' and 'WFPC2', the camera is selected by-chip and 
			should not be specified here.
		pos (list): Location of object of interest (spatial and spectral). [X, Y, chip, filter]
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		mask (bool): If True, uses data quality array to mask some pixels.
		writeto (bool): If True, will write 2D model PSF (differentiated with '_topsf' 
			suffix) and will amend relevant image WCS information/remove extraneous extensions.
		fov_arcsec (float): "Diameter" of model PSF image in arcsec.
		norm (float): Flux normalization for output PSF model.
		starselect (str): 'DAO', 'IRAF', or 'peak', which use DAOStarFinder, IRAFStarFinder, and 
			find_peaks from photutils respectively.
		starselectargs (dict): Keyword arguments for the chosen star detection method.
		thresh (float): Threshold for star detection (in multiples of standard deviations). 
			Note that the output ePSF is *incredibly* sensitive to this parameter, as it sets
			which stars are used to build the model.
		usermask (arr): If mask = True, used in addition to the DQ array to mask bad pixels. Must be
			the same dimensions as the data array.
		maskval (float): Value to assign to masked pixels. If None, default set by star seletion method. 
			Otherwise, can be np.nan or number. Default for 'IRAF' is np.nan, default for 'DAO' is 0.
		epsfargs (dict): Keyword arguments for the EPSFBuilder. Default in spike is to not oversample
			the PSF, but the regridding is all handled during the creation of the coord-specific model.

	Returns:
		ePSF model PSF
	
	"""

	ext = 1 #read data from relevant SCI extension
	extv = 1
	if (imcam in ['ACS/WFC', 'WFC3/UVIS']) and (pos[2] == 1):
		ext = 4
		extv = 2
	if (imcam in ['ACS/WFC', 'WFC3/UVIS']) and (pos[2] == 2):
		ext = 1 #yes, it's already 1, but this is to make things explicit
		extv = 1
	if imcam in ['WFPC', 'WFPC1', 'WFPC2']:
		ext = pos[2]
		extv = pos[2]


	coordstring = str(coords.ra)
	if coords.dec.deg >= 0:
		coordstring += '+'+str(coords.dec)
	if coords.dec.deg < 0:
		coordstring += str(coords.dec)

	modname = img.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_psf')


	pixkey = imcam #set pixel scale to get the dimensions of x, y
	if imcam in ['WFPC', 'WFPC1']:
		if pos[2] <= 4:
			pixkey += '_wf'
		if pos[2] >= 5:
			pixkey += '_pc'
	if (imcam == 'WFPC2') and (pos[2] == 1):
		pixkey += '_pc'
	if (imcam == 'WFPC2') and (pos[2] >= 2):
		pixkey += 'wf'
	if imcam == 'NIRCAM':
		if pos[2] in ['NRCA5', 'NRCB5', 'NRCALONG', 'NRCBLONG']:
			pixkey += '_long'
		if pos[2] not in ['NRCA5', 'NRCB5', 'NRCALONG', 'NRCBLONG']:
			pixkey += '_short'

	dat = fits.open(img)[ext].data

	mean, median, std = sigma_clipped_stats(dat, sigma=3.0)

	if starselect.upper() == 'DAO':
		# take default FWHM to be 4x the detector plate scale, can overwrite with starselectargs

		find = DAOStarFinder(threshold = thresh*std, **starselectargs)
		if not maskval:
			maskval = 0

	if starselect.upper() == 'IRAF':
		## from tests, nan masks work best with IRAF
		# suggested thresh = 3 here, as function of masking

		mean, median, std = sigma_clipped_stats(dat, sigma=3.0)

		find = IRAFStarFinder(threshold = thresh*std,  **starselectargs)
		if not maskval:
			maskval = np.nan	

	if verbose:
		print('Identifying stars to use in ePSF')
	
	if mask:
		maskarr = fits.open(img)[('DQ', extv)].data
		dat[maskarr > 0] = maskval # only retain good pixels
		maskarr[maskarr > 0] = True
		if usermask:
			dat[usermask] = maskval
			maskarr[usermask] = True
		sources = find(dat, maskarr)

	if not mask:
		sources = find(dat)

	exsize = int(2 * ((fov_arcsec/plate_scale[pixkey])//2) + 1) #size of extraction box
	xs = sources['xcentroid']
	ys = sources['ycentroid']
	exmask = ((xs > (exsize//2)) & (xs < (dat.shape[1] -1 - (exsize//2))) &
        (ys > (exsize//2)) & (ys < (dat.shape[0] -1 - (exsize//2))))
	tab = Table()
	tab['x'] = xs[exmask]
	tab['y'] = ys[exmask]
	nddata = NDData(data = dat - np.nanmedian(dat)) 
	if verbose:
		print('Beginning stellar extraction.')
	stars = extract_stars(nddata, tab, size = exsize)

	dimxy = fov_arcsec/plate_scale[pixkey] #make square PSF
	halfdim = dimxy//2
	xmin = int(pos[0]) - halfdim
	xmax = int(pos[0]) + halfdim
	ymin = int(pos[1]) - halfdim
	ymax = int(pos[1]) + halfdim

	x, y = np.meshgrid(np.arange(xmin, xmax+1), np.arange(ymin, ymax+1))

	if verbose:
		# ensure progress bar is toggled if verbose is true
		epsfargs['progress_bar'] = True

	if verbose:
		print('Creating ePSF builder.')
	epsfbuilder = EPSFBuilder(**epsfargs)

	if verbose:
		print('Starting PSF construction.')
	model, fitstars = epsfbuilder(stars)

	if verbose:
		print('Evaluating model at (%i, %i).'%(pos[0], pos[1]))
	psfmodel = model.evaluate(x = x, y = y, flux = norm, x_0 = int(pos[0]), y_0 = int(pos[1]))

	if plot:
		fig= plt.figure(figsize = (5, 5))
		plt.imshow(psfmodel, origin = 'lower', cmap = 'Greys', 
			vmin = np.nanpercentile(psfmodel, 20), vmax = np.nanpercentile(psfmodel, 97))
		plt.colorbar()
		fig.savefig(modname+'.png', bbox_inches = 'tight', dpi = 100)

		if verbose:
			print('PSF model image written to %s.png.'%(modname))

	if writeto:
		if verbose:
			print('Writing to %s.fits.'%modname.replace('_psf', '_topsf'))
		tools.rewrite_fits(psfmodel, coords, img, imcam, pos, method = 'ePSFs')

	return psfmodel


def psfex(coords, img, imcam, pos, plot = False, verbose = False, writeto = True, 
	savepsfex = False, seconf = None, psfconf = None, regrid = True, 
	mask = True, maskparams = {}):
	"""
	Generate PSFs using PSFEx.

	NOTE: For frames with fewer stars, this method is prone to artifacts, which is exacerbated by 
	using a lower detection threshold. (Also important to note that using the parallel option with 
	PSFEx may require additional attention to ensuring files do not overwrite each other given the
	default naming scheme for .cat and .psf files and subsequent steps to rename them.)

	Parameters:
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		imcam (str): Specification of instrument/camera used to capture the images (e.g., 'ACS/WFC', 'WFC3/IR', 'WFPC', 
			'WFPC2', 'MIRI', 'NIRCAM', 'NIRISS/Imaging'). For 'WFPC' and 'WFPC2', the camera is selected by-chip and 
			should not be specified here.
		img (str): Path to image for which PSF is generated.
		pos (list): [X, Y, chip, filter] as output from spike.tools.checkpixloc.
		plot (bool): If True, saves .pngs of the model PSFs.
		verbose (bool): If True, prints progress messages.
		writeto (bool): If True, will write 2D model PSF (differentiated with '_topsf' 
			suffix) and will amend relevant image WCS information/remove extraneous extensions. 
			This is in addition to the .psf file saved by PSFEx. No 2D PSF model is saved by PSFEx 
			by default, but this can be toggled in the tools.pypsfex arguments.
		savepsfex (str): If 'fits' or 'arr' save 2D model to that format.
		seconf (str): Path to SExtractor configuration file if not using default.
		peconf (str): Path to PSFEx configuration file if not using default.
		regrid (bool): If True, will (interpolate and) regrid model PSF to image pixel scale.
		mask (bool): If mask, apply mask generated by spike.tools.mask_fits to remove bad pixels.
		maskparams (dict): Any additional parameters to pass to spike.tools.mask_fits.

	Returns:
		2D PSFEx PSF model
	"""

	ext = 1 #run SExtractor only on relevant SCI extension
	extv = 1
	if (imcam in ['ACS/WFC', 'WFC3/UVIS']) and (pos[2] == 1):
		ext = 4
		extv = 2
	if (imcam in ['ACS/WFC', 'WFC3/UVIS']) and (pos[2] == 2):
		ext = 1 #yes, it's already 1, but this is to make things explicit
		extv = 1
	if imcam in ['WFPC', 'WFPC1', 'WFPC2']:
		ext = pos[2]
		extv = pos[2]

	if mask:
		tools.mask_fits(img, extv, **maskparams)
		if verbose:
			print('Finished masking data, running SExtractor')
		tools.pysextractor(img.replace('.fits', '_mask.fits')+'[%i]'%ext, config = seconf)	
	if not mask:
		if verbose:
			print('Running SExtractor')
		tools.pysextractor(img+'[%i]'%ext, config = seconf)
	if verbose:
		print('Finished SExtractor, running PSFEx')
	psfmodel = tools.pypsfex(img.replace('fits', 'cat'), pos, config = psfconf, savepsf = savepsfex)
	if verbose:
		print('Finished PSFEx, generating image')

	coordstring = str(coords.ra)
	if coords.dec.deg >= 0:
		coordstring += '+'+str(coords.dec)
	if coords.dec.deg < 0:
		coordstring += str(coords.dec)

	modname = img.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_psf')

	if plot:
		fig= plt.figure(figsize = (5, 5))
		plt.imshow(psfmodel, origin = 'lower', cmap = 'Greys', 
			vmin = np.nanpercentile(psfmodel, 20), vmax = np.nanpercentile(psfmodel, 97))
		plt.colorbar()
		fig.savefig(modname+'.png', bbox_inches = 'tight', dpi = 100)

		if verbose:
			print('PSF model image written to %s.png'%(modname))

	if writeto:
		if verbose:
			print('Writing to %s.fits.'%modname.replace('_psf', '_topsf'))
		tools.rewrite_fits(psfmodel, coords, img, imcam, pos, method = 'PSFEx')

	return psfmodel


	




"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/lib/suffix.py
"""

"""Suffix manipulation

API
---
`remove_suffix`
    Main function used to remove suffixes from
    file base names.

Notes
-----

`KNOW_SUFFIXES` is the list used by `remove_suffix`. This
list is generated by the function `combine_suffixes`. The function uses
`SUFFIXES_TO_ADD` to add other suffixes that it would otherwise not
discover or should absolutely be in the list. The function uses
'SUFFIXES_TO_DISCARD` for strings found that are not to be considered
suffixes.

Hence, to update `KNOW_SUFFIXES`, update both `SUFFIXES_TO_ADD` and
`SUFFIXES_TO_DISCARD` as necessary, then use the output of
`find_suffixes`.
"""

import itertools
import logging
import re

__all__ = ["remove_suffix"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Suffixes that are hard-coded or otherwise
# have to exist. Used by `find_suffixes` to
# add to the result it produces.
SUFFIXES_TO_ADD = [
    "ca",
    "crf",
    "cal",
    "dark",
    "flat",
    "median",
    "phot",
    "photom",
    "ramp",
    "rate",
    "uncal",
    "assignwcs",
    "dq_init",
    "dqinit",
    "assign_wcs",
    "linearity",
    "jump",
    "rampfit",
    "saturation",
    "dark_current",
    "darkcurrent",
    "outlier_detection",
    "skymatch",
    "refpix",
    "cat",
    "segm",
    "coadd",
]

# Suffixes that are discovered but should not be considered.
# Used by `find_suffixes` to remove undesired values it has found.
SUFFIXES_TO_DISCARD = ["mosaicpipeline", "pipeline", "step"]


# Calculated suffixes.
_calculated_suffixes = {
    "saturationstep",
    "darkcurrentstep",
    "jumpstep",
    "rampfit",
    "dark_current",
    "assignwcsstep",
    "flatfieldstep",
    "fluxstep",
    "step",
    "dqinitstep",
    "assign_wcs",
    "linearity",
    "rampfitstep",
    "photomstep",
    "pipeline",
    "dq_init",
    "linearitystep",
    "dark_current",
    "jump",
    "tweakregstep",
    "outlierdetectionstep",
    "skymatchstep",
    "refpixstep",
    "resamplestep",
    "sourcecatalogstep",
    "multibandcatalogstep",
}


# ##########
# Suffix API
# ##########
def remove_suffix(name):
    """Remove the suffix if a known suffix is already in name"""
    separator = None
    match = REMOVE_SUFFIX_REGEX.match(name)
    try:
        name = match.group("root")
        separator = match.group("separator")
    except AttributeError:
        pass
    if separator is None:
        separator = "_"
    return name, separator


def replace_suffix(name, new_suffix):
    """Replace suffix on name

    Parameters
    ----------
    name: str
        The name to replace the suffix of.
        Expected to be only the basename; no extensions.

    new_suffix:
        The new suffix to use.
    """
    no_suffix, separator = remove_suffix(name)
    return no_suffix + separator + new_suffix


# #####################################
# Functions to generate `KNOW_SUFFIXES`
# #####################################
def combine_suffixes(
    to_add=(_calculated_suffixes, SUFFIXES_TO_ADD), to_remove=(SUFFIXES_TO_DISCARD,)
):
    """Combine the suffix lists into a single list

    Parameters
    ----------
    to_add: [iterable[, ...]]
        List of iterables to add to the combined list.

    to_remove: [iterable[, ...]]
        List of iterables to remove from the combined list.

    Returns
    -------
    suffixes: list
        The list of suffixes.
    """
    combined = set(itertools.chain.from_iterable(to_add))
    combined.difference_update(itertools.chain.from_iterable(to_remove))
    combined = list(combined)
    combined.sort()

    return combined


def find_suffixes():
    """Find all possible suffixes from the romancal package

    Returns
    -------
    suffixes: set
        The set of all programmatically findable suffixes.

    Notes
    -----
    This will load all of the `romancal` package. Consider if this
    is worth doing dynamically or only as a utility to update
    a static list.
    """
    from spike.romancal.stpipe_utilities import all_steps

    # First traverse the code base and find all
    # `Step` classes. The default suffix is the
    # class name.
    suffixes = {klass_name.lower() for klass_name, klass in all_steps().items()}

    # That's all folks
    return list(suffixes)


# --------------------------------------------------
# The set of suffixes used by the pipeline.
# This set is generated by `combine_suffixes`.
# Only update this list by `combine_suffixes`.
# Modify `SUFFIXES_TO_ADD` and `SUFFIXES_TO_DISCARD`
# to change the results.
# --------------------------------------------------
KNOW_SUFFIXES = combine_suffixes()

# Regex for removal
REMOVE_SUFFIX_REGEX = re.compile(
    "^(?P<root>.+?)((?P<separator>_|-)(" + "|".join(KNOW_SUFFIXES) + "))?$"
)


# ############################################
# Main
# Find and report differences from known list.
# ############################################
if __name__ == "__main__":
    print("Searching code base for calibration suffixes...")
    calculated_suffixes = find_suffixes()
    found_suffixes = combine_suffixes(
        to_add=(calculated_suffixes, SUFFIXES_TO_ADD), to_remove=(SUFFIXES_TO_DISCARD,)
    )
    print(
        "Known list has {known_len} suffixes. Found {new_len} suffixes.".format(
            known_len=len(KNOW_SUFFIXES), new_len=len(found_suffixes)
        )
    )
    print(
        "Suffixes that have changed are"
        f" {set(found_suffixes).symmetric_difference(KNOW_SUFFIXES)}"
    )


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/format_template.py
"""

from collections import defaultdict
from string import Formatter

__all__ = ["FormatTemplate"]


# Define conversion based on format type
CONVERSION = {
    "%": float,
    "b": int,
    "c": int,
    "d": int,
    "e": float,
    "E": float,
    "f": float,
    "F": float,
    "g": float,
    "G": float,
    "n": int,
    "o": int,
    "s": str,
    "x": int,
    "X": int,
}


class FormatTemplate(Formatter):
    """Format a template

    Parameters
    ----------
    template : str
        A Python format string

    separator : string
        Separator to use for values which have no
        matching replacement strings

    key_formats : dict or None
        A dict of key-specific formatting where the value will
        be pre-formatted before being passed to the final format
        string. Each format will be tried until success.

    remove_unused : bool
        By default, unused replacement fields are left in the
        result, for use in subsequent replacement usage.
        If True, such fields are removed from the result.

    kwargs : dict or named parameters
        The key/values pairs to fill into the Python format string

    Returns
    -------
    str
        The formatted string

    Notes
    -----
    This differences from Pythons `format` method are:
        - If a replacement field does not have a given value,
          the replacement field is left in the result
        - If a key/value pair is present but has no replacement field,
          the value is simply appended.
        - Template can only use named replacement fields.

    Examples
    --------
    The basic example:
    >>> template = 'name="{name}" value="{value}"'
    >>> fmt = FormatTemplate()
    >>> fmt(template)
    'name="{name}" value="{value}"'

    But with actual values given:
    >>> fmt(template, name='fred', value='great')
    'name="fred" value="great"'

    But wait, too many values given:
    >>> fmt(template, name='fred', value='great', extra='more')
    'name="fred" value="great"_more'

    And with too many and not enough:
    >>> fmt(template, value='great', extra='more')
    'name="{name}" value="great"_more'

    With a different separator:
    >>> fmt.separator = '---'
    >>> fmt(template, name='fred', value='great', extra='more')
    'name="fred" value="great"---more'

    Initializing with a different separator:
    >>> fmt_newsep = FormatTemplate(separator='_now-with_')
    >>> fmt_newsep(template, name='fred', value='great', extra='more')
    'name="fred" value="great"_now-with_more'

    Setup preformatting
    >>> key_formats = {'value': ['pre_{:s}_format']}
    >>> fmt_preformat = FormatTemplate(key_formats=key_formats)
    >>> fmt_preformat(template, name='fred', value='great')
    'name="fred" value="pre_great_format"'
    """

    def __init__(self, separator="_", key_formats=None, remove_unused=False):
        """Initialize class

        Parameters
        ----------
        separator : str
            For key/value pairs given that do not have a
            replacement field, the values are appended to
            the string using this separator.

        key_formats : {key: format(, ...)}
            dict of formats to pre-format the related values
            before insertion into the template.
        """
        super().__init__()
        self.separator = separator
        self.remove_unused = remove_unused
        self._used_keys = []

        self.key_formats = defaultdict(lambda: ["{:s}"])
        if key_formats:
            self.key_formats.update(key_formats)

    def format(self, format_string, **kwargs):
        """Perform the formatting

        Parameters
        ----------
        format_string : str
            The string to be formatted

        kwargs : dict
            The key/value pairs to insert into the string

        Returns
        -------
        formatted : str
            The formatted string.
        """
        self._used_keys = []

        # Preformat the values
        formatted_kwargs = {}
        for key, value in kwargs.items():
            if value is not None:
                for key_format in self.key_formats[key]:
                    # Get the formatting type character. Indices are:
                    #  0: The first replacement field. There should only be one.
                    #  2: Get the format spec.
                    #  -1: Get the last character representing the type.
                    format_type = next(iter(self.parse(key_format)))[2][-1]

                    try:
                        value = key_format.format(CONVERSION[format_type](value))
                    except ValueError:
                        pass
                    else:
                        break
                else:
                    raise RuntimeError(
                        f"No suitable formatting for {key}: {value} found. Given"
                        f" formatting options:\n\t{self.key_formats[key]}"
                    )
            formatted_kwargs[key] = value
        result = super().format(format_string, **formatted_kwargs)

        # Get any unused arguments and simply do the appending
        unused_keys = set(formatted_kwargs).difference(self._used_keys)
        unused_values = [
            formatted_kwargs[unused]
            for unused in unused_keys
            if formatted_kwargs[unused] is not None
        ]
        result_parts = [result, *unused_values]
        return self.separator.join(result_parts)

    # Make the instance callable
    __call__ = format

    def get_value(self, key, args, kwargs):  # noqa: ARG002
        """Return a given field value

        Parameters
        ----------
        key : str
            The key to retrieve.

        args : [arg(, ...)]
            Positional arguments passed.
            This is ignored.

        kwargs : {k:v(, ...)}
            The key/value pairs passed in.

        Returns
        -------
        obj
            The value from the kwargs.
            If not found, the string '{key}' is returned.
        """
        if self.remove_unused:
            default = ""
        else:
            default = "{" + key + "}"
        value = kwargs.get(key, default)
        self._used_keys.append(key)
        if value is None:
            value = ""

        return value


from asdf import AsdfFile
from astropy.coordinates import SkyCoord, name_resolve
import astropy
import astropy.units as u
from astropy.io import fits
from astropy.wcs import WCS, utils
import numpy as np
import os
import pkg_resources
from scipy.interpolate import RectBivariateSpline
import warnings

def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
	return '%s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)

warnings.formatwarning = warning_on_one_line

CONFIG_PATH = pkg_resources.resource_filename('spike', 'configs/')

# #########
#  * * * * 
# #########

def objloc(obj):
	"""
	Get object location.

	Parameters:
		obj (str): Name or coordinates for object of interest. If coordinates, should be in
			HH:MM:SS DD:MM:SS or degree formats. Names must be resolvable in SIMBAD.
	Returns:
		coords (astropy coordinates object)
	"""
	isname = False #check if obj is name or coordinates
	for s in obj:
		if s.isalpha():
			isname = True
			break

	if isname:
		coords = name_resolve.get_icrs_coordinates(obj)

	if not isname:
		if ':' in obj:
			coords = SkyCoord(obj, unit = (u.hour, u.deg), frame = 'icrs')
		if not ':' in obj:
			coords = SkyCoord(obj, unit = u.deg, frame = 'icrs')

	return coords


def checkpixloc(coords, img, inst, camera = None):
	"""
	Get object location on detector. 

	Parameters:
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		img (str): Path to image.
		inst (str): Instrument of interest. 
				HST: 'ACS', 'WFC3', 'WFPC', WFPC2', 'NICMOS'
				JWST: 'MIRI', 'NIRCAM', 'NIRISS'
				Roman: 'WFI', 'CGI'
		camera (str): Camera associated with instrument.
				HST/ACS: 'WFC', 'HRC'
				HST/WFC3: 'UVIS', 'IR'
				JWST/NIRISS: 'Imaging', 'AMI' #AMI has different multi-extension mode

	Returns:
		[X, Y, chip, filter] (list): Pixel coordinates, chip number (HST) or detector name (JWST/Roman) if relevant, and filter name.
			Only returned if object coordinates fall onto detector - returns NaNs if not.

	"""
	hdu = fits.open(img)
	if camera:
		imcam = inst.upper() + '/' + camera.upper()
	if not camera:
		imcam = inst.upper()

	try: #get filter
		filt = hdu[0].header['FILTER']
	except:
		if hdu[0].header['FILTER1'].startswith('F'):
			filt = hdu[0].header['FILTER1']
		else:
			filt = hdu[0].header['FILTER2']

	### instrument checks ###
	if imcam in ['ACS/WFC', 'WFC3/UVIS']:
		chip1 = hdu[4]
		chip2 = hdu[1]
		chips = [chip1, chip2]

		chip = np.nan
		for a in chips:
			wcs1 = WCS(a.header, fobj = hdu)
			datshape = a.data.shape[::-1] #transposed based on numpy vs. fits preference
			if type(coords) != astropy.coordinates.sky_coordinate.SkyCoord:
				xcoord_out = []
				ycoord_out = []
				chip_out = []
				for coord in coords:
					check = utils.skycoord_to_pixel(coord, wcs1)
					if np.logical_and(0 <= check[0] <= datshape[0], 0 <= check[1] <= datshape[1]):
						xcoord_out.append(check[0])
						ycoord_out.append(check[1])
						if a == chip1:
							chip_out.append('1')
						if a == chip2:
							chip_out.append('2')
				if len(xcoord_out) >= 1:
					out = [[xcoord_out[i], ycoord_out[i], chip_out[i], filt] for i in range(len(coords))]
				if len(xcoord_out) == 0:
					out = [np.nan] * 4
			if type(coords) == astropy.coordinates.sky_coordinate.SkyCoord:
				check = utils.skycoord_to_pixel(coords, wcs1)
				if np.logical_and(0 <= check[0] <= datshape[0], 0 <= check[1] <= datshape[1]):
					x_coord = check[0]
					y_coord = check[1]
					if a == chip1:
						chip = 1
					if a == chip2:
						chip = 2
		if (type(chip) != str) and (np.isnan(chip)):
			out = [np.nan, np.nan, chip, np.nan]
		else:
			out = [float(x_coord), float(y_coord), chip, filt]

	if imcam in ['WFPC', 'WFPC1', 'WFPC2']:
		if imcam in ['WFPC', 'WFPC1']: #accounting for use of both names
			chip1 = hdu[1]
			chip2 = hdu[2]
			chip3 = hdu[3]
			chip4 = hdu[4]
			chip1 = hdu[5]
			chip2 = hdu[6]
			chip3 = hdu[7]
			chip4 = hdu[8]
			chips = [chip1, chip2, chip3, chip4, chip5, chip6, chip7, chip8]
		if imcam == 'WFPC2':
			chip1 = hdu[1]
			chip2 = hdu[2]
			chip3 = hdu[3]
			chip4 = hdu[4]
			chips = [chip1, chip2, chip3, chip4]

		chip = np.nan
		for a in chips:
			wcs1 = WCS(a.header, fobj = hdu)
			datshape = a.data.shape[::-1] #transposed based on numpy vs. fits preference
			if type(coords) != astropy.coordinates.sky_coordinate.SkyCoord:
				xcoord_out = []
				ycoord_out = []
				chip_out = []
				for coord in coords:
					check = utils.skycoord_to_pixel(coord, wcs1)
					if np.logical_and(0 <= check[0] <= datshape[0], 0 <= check[1] <= datshape[1]):
						xcoord_out.append(check[0])
						ycoord_out.append(check[1])
						if a == chip1:
							chip_out.append('1')
						if a == chip2:
							chip_out.append('2')
				if len(xcoord_out) >= 1:
					out = [[float(xcoord_out[i]), float(ycoord_out[i]), chip_out[i], filt] for i in range(len(coords))]
				if len(xcoord_out) == 0:
					out = [np.nan] * 4
			if type(coords) == astropy.coordinates.sky_coordinate.SkyCoord:
				check = utils.skycoord_to_pixel(coords, wcs1)
				if np.logical_and(0 <= check[0] <= datshape[0], 0 <= check[1] <= datshape[1]):
					x_coord = check[0]
					y_coord = check[1]

					chip = np.where(np.array(chips) == a)[0][0] + 1
					
		if (type(chip) != str) and (np.isnan(chip)):
			out = [np.nan, np.nan, chip, np.nan]
		else:
			out = [float(x_coord), float(y_coord), chip, filt]


	if imcam in ['ACS/HRC', 'WFC3/IR', 'NICMOS',
				 'MIRI', 'NIRCAM', 'NIRISS', 'NIRISS/IMAGING', 
				 'WFI', 'CGI']:
		# for WFC3, only checks the final readout by design
		chip = 0 #no chip
		wcs1 = WCS(hdu[1].header, fobj = hdu)
		datshape = hdu[1].data.shape[::-1] #transposed based on numpy vs. fits preference
		if type(coords) != astropy.coordinates.sky_coordinate.SkyCoord:
			xcoord_out = []
			ycoord_out = []
			chip_out = []
			for coord in coords:
				check = utils.skycoord_to_pixel(coord, wcs1)
				xcoord_out.append(check[0])
				ycoord_out.append(check[1])
				if imcam == 'NIRCAM':
					chip = hdu[0].header['DETECTOR']
					if chip in ['NRCALONG', 'NRCBLONG']:
						chip = chip.replace('LONG', '5')
				if imcam == 'WFI':
					# based on how SCA detector is identified in simulated data: https://roman.ipac.caltech.edu/sims/Simulations_csv.html
					strlist = img.split('_')
					chip = 'SCA%s'%strlist[-1].split('.')[0].rjust(2, '0')
				chip_out.append(chip)
			if len(xcoord_out) >= 1:
				out = [[float(xcoord_out[i]), float(ycoord_out[i]), chip_out[i], filt] for i in range(len(coords))]
			if len(xcoord_out) == 0:
				out = [np.nan] * 4


		if type(coords) == astropy.coordinates.sky_coordinate.SkyCoord:
			check = utils.skycoord_to_pixel(coords, wcs1)
			if np.logical_and(0 <= check[0] <= datshape[0], 0 <= check[1] <= datshape[1]):
				x_coord = check[0]
				y_coord = check[1]
				if imcam == 'NIRCAM':
					chip = hdu[0].header['DETECTOR']
					if chip in ['NRCALONG', 'NRCBLONG']:
						chip = chip.replace('LONG', '5')
				if imcam == 'WFI':
					# based on how SCA detector is identified in simulated data: https://roman.ipac.caltech.edu/sims/Simulations_csv.html
					strlist = img.split('_')
					chip = 'SCA%s'%strlist[-1].split('.')[0].rjust(2, '0')
				out = [float(x_coord), float(y_coord), chip, filt]
			else:
				out = [np.nan] * 4

	return out

	#add support for FOC and NIRISS AMI


def to_asdf(fitspath, save = True):
	"""
	Convert .fits file to .asdf by simply wrapping data and header extensions.

	Parameters:
		fitspath (str): Path to .fits file to convert.
		save (bool): If True, saves the .asdf file.

	Returns:
		ASDF file object
	"""

	fits_in = fits.open(fitspath)
	# dicts = [{} for i in range(len(orig))]
	headers = {}
	for i in range(len(fits_in)):
		headers['head'+str(i)] = {**fits_in[i].header}
	
	asdf_out = AsdfFile(headers)

	for i in range(len(fits_in)):
		asdf_out['dat'+str(i)] = fits_in[i].data

	if save:
		asdf_out.write_to(fitspath.replace('fits', 'asdf'))

	return asdf_out


def pysextractor(img_path, config = None, psf = True, userargs = None, keepconfig = False):
	"""
	Wrapper to easily call SExtractor from python. 
	Using this in lieu of sep for easier compatibility with PSFEx.

	Parameters:
		img_path (str): Path to the image from the working directory.
		config (str): If specifying custom config, path to config file. If none, uses default.sex.
		psf (bool): If True and no configuration file specified, uses config and param files that work with PSFEx.
		userargs (str): Any additional command line arguments to feed to SExtractor. The preferred way to include
			user arguments is via specification in the config file as command line arguments simply override the 
			corresponding configuration setting.
		keepconfig (str): If True, retain parameter files and convolutional kernels moved to working dir.

	Returns:
		Generates a .cat file with the same name as img_path

	"""

	if not config:
		if psf:
			configpath = CONFIG_PATH + 'sextractor_config/default_psf.sex'
		if not psf:
			configpath = CONFIG_PATH + 'sextractor_config/default.sex'
		os.system('cp '+ CONFIG_PATH +'sextractor_config/default.conv .')
		os.system('cp '+ CONFIG_PATH +'sextractor_config/* .')
	if config:
		configpath = config

	sextractor_args = 'sex '+img_path+' -c '+configpath

	if userargs:
		sextractor_args += ' '+userargs

	os.system(sextractor_args)

	imgpath = img_path.split('[')[0]
	if imgpath.split('_')[-1].startswith('mask'):
		imgpath = imgpath.replace('_mask.fits', '.fits')
	os.system('mv test.cat ' + imgpath.replace('fits', 'cat')) #move to img name

	if (not keepconfig) and (not config):
		# clean up user directory by removing copied files
		os.system('rm default*')
		os.system('rm *.conv')


def regridarr (im, sample):
	"""
	Regrid PSF model to input pixel scale.

	Parameters:
		im (arr): PSF image array.
		sample (float): Npix,im/Npix,orig. If sample > 1, oversampled; if sample < 1, undersampled.

	Returns:
		Interpolated and regridded PSF model.

	"""

	if sample == 1.:
		return im

	x,y = im.shape
	xnew = np.arange(0, x, sample)
	ynew = np.arange(0, y, sample)

	spline = RectBivariateSpline(np.arange(x), np.arange(y), im)
	out = spline(xnew[xnew <= x-1], ynew[ynew <= y-1])

	return out	


def psfexim(psfpath, pixloc, regrid = True, save = False):
	"""
	Generate image from PSFEx .psf file.

	Parameters:
		psfpath (str): Path to the relevant .psf file from the working directory.
		pixloc (tuple): Pixel location of object of interest in (x, y).
		regrid (bool): If True, will (interpolate and) regrid model PSF to image pixel scale.
		save (str): If 'fits' or 'arr', will save in the specified format with name from psfpath.
			The option to save as an array results in a .npy file.

	Returns:
		2D image of PSFEx model
	"""
	psfexmodel = fits.open(psfpath)

	if psfexmodel[1].header['PSFAXIS3'] == 1:
		warnings.warn('PSF model is based on only a single component vector.', Warning, stacklevel = 2)
		psfmodel = psfexmodel[1].data['PSF_MASK'][0, 0, :, :]

	else:
		x_, y_ = pixloc

		x = (x_ - psfexmodel[1].header['POLZERO1'])/psfexmodel[1].header['POLSCAL1']
		y = (y_ - psfexmodel[1].header['POLZERO2'])/psfexmodel[1].header['POLSCAL2']

		order = psfexmodel[1].header['POLDEG1']

		xpoly, ypoly = x**np.arange(order+1), y**np.arange(order+1)

		# takes X_c * Phi_c, where Phi is the vector and X_c(x, y) is the basis function
		# see https://psfex.readthedocs.io/en/latest/Working.html and
		# https://www.astromatic.net/wp-content/uploads/psfex_article.pdf

		xc = []
		for i, yy in enumerate(ypoly):
			for ii, xx in enumerate(xpoly[:(order+1-i)]):
				xc.append(xx*yy)

		phic = psfexmodel[1].data['PSF_MASK'][0]

		psfmodel = np.sum(phic * np.array(xc)[:, None, None], axis = 0)

	if regrid:
		if psfexmodel[1].header['PSF_SAMP'] != 1.:
			psfmodel = regridarr(psfmodel, psfexmodel[1].header['PSF_SAMP'])

	if save:
		if save.lower() not in ['arr', 'fits']:
			warnings.warn('Save input must be "arr" or "fits". Generating PSF model without saving.', Warning, stacklevel = 2)
		if save.lower() == 'arr':
			np.save(psfpath.replace('.psf', '_psfex_psf.npy'), psfmodel)
		if save.lower() == 'fits':
			fits.writeto(psfpath.replace('.psf', '_psfex_psf.fits'), psfmodel)

	return psfmodel


def pypsfex(cat_path, pos, config = None, userargs = None, makepsf = True, 
	savepsf = False, keepconfig = False, regrid = True):
	"""
	Wrapper to easily call PSFEx from python. 

	Parameters:
		cat_path (str): Path to the SExtractor catalog from the working directory.
		pos(tuple): Pixel (and spectral) location of object of interest in (x, y, chip, filter) - as from spike.tools.checkpixloc.
			If not specified, generates generic model assuming central pixel.
		config (str): If specifying custom config, path to config file. If none, uses default.psfex.
		userargs (str): Any additional command line arguments to feed to PSFEx. The preferred way to include
			user arguments is via specification in the config file as command line arguments simply override the 
			corresponding configuration setting.
		makepsf (bool): If True, returns 2D PSF model.
		savepsf (str): If 'fits', 'arr', or 'txt', will save 2D model PSF in that file format with the same name as the catalog.
		keepconfig (str): If True, retain parameter files and convolutional kernels moved to working dir.
		regrid (bool): If True, will (interpolate and) regrid model PSF to image pixel scale.

	Returns:
		Generates a .psf file that stores linear bases of PSF.
		If makepsf = True, also returns 2D array containing PSF model.
	"""

	if not config:
		configpath = CONFIG_PATH + 'psfex_config/default.psfex'
		os.system('cp '+ CONFIG_PATH +'psfex_config/* .')
	if config:
		configpath = config

	psfex_args = 'psfex '+cat_path+' -c '+configpath

	if userargs:
		psfex_args += ' '+userargs

	os.system(psfex_args)
	os.system('mv test.psf ' + cat_path.replace('cat', 'psf')) #move to img name

	if (not keepconfig) and (not config):
		# clean up user directory by removing copied files
		os.system('rm default*')
		os.system('rm *.conv')

	if makepsf:

		if pos:
			x, y, chip, filts = pos

		if not pos:
			#assumes that .cat and .fits file are in the same directory
			im = fits.open(cat_path.replace('cat', 'fits'))[1].data
			x, y = im.shape/2 #select central pixel if not specified


		psfmodel = psfexim(cat_path.replace('cat', 'psf'), pixloc = (x, y), save = savepsf, regrid = regrid)
			
		return psfmodel


def rewrite_fits(psfarr, coords, img, imcam, pos, method = None):
	"""
	Write relevant image headers to the model PSFs and modify the coordinates and WCS.
	Creates a full _topsf.fits file with only one SCI extension for use with drizzle/resample.

	Parameters:
		psfarr (arr): The 2D PSF model.
		coords (astropy skycoord object): Coordinates of object of interest or list of skycoord objects.
		img (str): Path to image for which PSF is generated.
		imcam (str): 'ACS/WFC' is the only recommended instrument/camera combination for this PSF generation method.
		pos (list): Location of object of interest (spatial and spectral).[X, Y, chip, filter]
		method (list): Method used to generate PSF.

	Returns: 
		Generates a new FITS file with a _topsf suffix, which stores the 2D PSF model in the 
		'SCI' extension and inherits the header information from the original image.

	"""

	ext = 1
	extv = 1
	if (imcam in ['ACS/WFC', 'WFC3/UVIS']) and (pos[2] == 1):
		ext = 4
		extv = 2
	if (imcam in ['ACS/WFC', 'WFC3/UVIS']) and (pos[2] == 2):
		ext = 1 #yes, it's already 1, but this is to make things explicit
		extv = 1
	if imcam in ['WFPC', 'WFPC1', 'WFPC2']:
		ext = pos[2]
		extv = pos[2]

	imgdat = fits.open(img)

	psfim = np.zeros_like(imgdat[ext].data)
	xmin = int(pos[0]) - psfarr.shape[0]//2
	xmax = int(pos[0]) + psfarr.shape[0]//2
	ymin = int(pos[1]) - psfarr.shape[1]//2
	ymax = int(pos[1]) + psfarr.shape[1]//2

	if psfarr.shape[0] % 2 == 0:
		psfim[ymin:ymax, xmin:xmax] += psfarr
	if psfarr.shape[0] % 2 != 0:
		psfim[ymin:ymax+1, xmin:xmax+1] += psfarr

	cphdr = fits.PrimaryHDU(header = imgdat[0].header)

	hdr = imgdat[ext].header
	if method:
		hdr['COMMENT'] = "PSF generated using %s via spike."%method
	if not method:
		hdr['COMMENT'] = "PSF generated via spike."
	cihdr = fits.ImageHDU(data = psfim, header = hdr, name = 'SCI', ver = 1)

	if img.split('_')[-1] != '_c0m.fits':
		ehdrdat = np.zeros_like(imgdat[('ERR', extv)].data) #shouldn't matter, but doing this explicitly anyway
		dqhdrdat = np.zeros_like(imgdat[('DQ', extv)].data)
		cehdr = fits.ImageHDU(data = ehdrdat, header = imgdat[('ERR', extv)].header, name = 'ERR', ver = 1)
		cdqhdr = fits.ImageHDU(data = dqhdrdat, header = imgdat[('DQ', extv)].header, name = 'DQ', ver = 1)

	coordstring = str(coords.ra)
	if coords.dec.deg >= 0:
		coordstring += '+'+str(coords.dec)
	if coords.dec.deg < 0:
		coordstring += str(coords.dec)

	modname = img.replace('.fits', '_'+coordstring+'_%s'%pos[3]+'_topsf.fits')

	if img.split('_')[-1] != '_c0m.fits':
		hdlist = [cphdr, cihdr, cehdr, cdqhdr]

	if img.split('_')[-1] == '_c0m.fits':
		hdlist = [cphdr, cihdr]

	try: #get WCSDVARR
		dp1 = hdr['DP1']
	except:
		dp1 = None

	try: #get WCSDVARR
		dp2 = hdr['DP2']
	except:
		dp2 = None

	try: #get D2IMARR
		d2im1 = hdr['D2IM1']
	except:
		d2im1 = None

	try: #get D2IMARR
		d2im2 = hdr['D2IM2']
	except:
		d2im2 = None


	if (d2im1 in ['EXTVER: 1', 'EXTVER: 1.0']) & (d2im2 in ['EXTVER: 2', 'EXTVER: 2.0']):
		hdlist.append(fits.ImageHDU(data = imgdat[('D2IMARR', 1)].data, header = imgdat[('D2IMARR', 1)].header, 
			name = 'D2IMARR', ver = 1))
		hdlist.append(fits.ImageHDU(data = imgdat[('D2IMARR', 2)].data, header = imgdat[('D2IMARR', 2)].header, 
			name = 'D2IMARR', ver = 2))
	if (d2im1 in ['EXTVER: 3', 'EXTVER: 3.0']) & (d2im2 in ['EXTVER: 4', 'EXTVER: 4.0']):
		hdlist.append(fits.ImageHDU(data = imgdat[('D2IMARR', 3)].data, header = imgdat[('D2IMARR', 3)].header, 
			name = 'D2IMARR', ver = 3))
		hdlist.append(fits.ImageHDU(data = imgdat[('D2IMARR', 4)].data, header = imgdat[('D2IMARR', 4)].header, 
			name = 'D2IMARR', ver = 4))

	if (dp1 in ['EXTVER: 1', 'EXTVER: 1.0']) & (dp2 in ['EXTVER: 2', 'EXTVER: 2.0']):
		hdlist.append(fits.ImageHDU(data = imgdat[('WCSDVARR', 1)].data, header = imgdat[('WCSDVARR', 1)].header, 
			name = 'WCSDVARR', ver = 1))
		hdlist.append(fits.ImageHDU(data = imgdat[('WCSDVARR', 2)].data, header = imgdat[('WCSDVARR', 2)].header, 
			name = 'WCSDVARR', ver = 2))
	if (dp1 in ['EXTVER: 3', 'EXTVER: 3.0']) & (dp2 in ['EXTVER: 4', 'EXTVER: 4.0']):
		hdlist.append(fits.ImageHDU(data = imgdat[('WCSDVARR', 3)].data, header = imgdat[('WCSDVARR', 3)].header, 
			name = 'WCSDVARR', ver = 3))
		hdlist.append(fits.ImageHDU(data = imgdat[('WCSDVARR', 4)].data, header = imgdat[('WCSDVARR', 4)].header, 
			name = 'WCSDVARR', ver = 4))

	if imcam in ['NIRCAM', 'MIRI', 'NIRISS']:
		hdlist.append(fits.ImageHDU(data = imgdat['AREA', 1].data, header = imgdat['AREA', 1].header))
		hdlist.append(fits.ImageHDU(data = imgdat['VAR_POISSON', 1].data, header = imgdat['VAR_POISSON', 1].header))
		hdlist.append(fits.ImageHDU(data = imgdat['VAR_RNOISE', 1].data, header = imgdat['VAR_RNOISE', 1].header))
		hdlist.append(fits.ImageHDU(data = imgdat['VAR_FLAT', 1].data, header = imgdat['VAR_FLAT', 1].header))
		hdlist.append(fits.BinTableHDU(data = imgdat['ASDF', 1].data, header = imgdat['ASDF', 1].header))

	hdulist = fits.HDUList(hdlist)

	if img.split('_')[-1] != '_c0m.fits':
		hdulist.writeto(modname)

	if img.split('_')[-1] == '_c0m.fits':
		modname = modname.replace('_topsf.fits', '_topsf_c0m.fits')
		hdulist.writeto(modname)
		os.system('cp %s %s'%(img.replace('c0m.fits', 'c1m.fits'), modname.replace('c0m.fits', 'c1m.fits')))


def mask_fits(img, ext = 1, maskdq = True, dqthresh = 0, 
	maskerr = False, errthresh = 20, usermask = None, fillval = 0):
	"""
	Generate a FITS file that fills in masked pixels with a specified value. Useful for
	feeding to e.g., SExtractor. Preserves truncated FITS extension structure, for the specified
	extension on which the the mask was applied plus ERR and DQ extensions. 
	(Assumes that extensions are named SCI, ERR, DQ.)

	Parameters:
		img (str): Path to image to which mask is applied.
		ext (int): Integer index for *version* of extension to which mask should be applied.
			This will be used to index dat[('SCI', ext)] etc.
		maskdq (bool): If True, masks values above dqthresh.
		dqthresh (float): Maximum acceptable value in DQ image. Pixels above dqthresh will be masked.
		maskerr (bool): If True, masks values above errthresh.
		errthresh (float): Maximum acceptable value in ERR image. Pixels above errthresh will be masked.
		usermask (arr): If specified, used as mask on data array (good pixels should have value of 0). 
			Must be the same dimensions as the data array. (Can be used in addition to DQ and ERR masking.)
		fillval (float): Value with which to fill the masked pixels.

	Returns: 
		Generates a new FITS file with a _mask suffix with masked pixels filled in by fillval.

	"""

	imgdat = fits.open(img)

	cphdr = fits.PrimaryHDU(header = imgdat[0].header)

	hdr = imgdat[('SCI', ext)].header
	dat = imgdat[('SCI', ext)].data

	if img.split('_')[-1] == '_c0m.fits':

		errdat = fits.open(img.split('_')[:-1]+'_c1m.fits')

		dq = errdat[('DQ', ext)].data
		err = errdat[('ERR', ext)].data

		if maskdq:
			dat[dq > dqthresh] = fillval
		if maskerr:
			dat[err > errthresh] = fillval
		if usermask:
			dat[usermask > 0] = fillval

		cihdr = fits.ImageHDU(data = dat, header = hdr, name = 'SCI')
		cehdr = fits.ImageHDU(data = err, header = errdat[('ERR', ext)].header, name = 'ERR')
		cdqhdr = fits.ImageHDU(data = dq, header = errdat[('DQ', ext)].header, name = 'DQ')


	else:

		dq = imgdat[('DQ', ext)].data
		err = imgdat[('ERR', ext)].data

		if maskdq:
			dat[dq > dqthresh] = fillval
		if maskerr:
			dat[err > errthresh] = fillval
		if usermask:
			dat[usermask > 0] = fillval

		cihdr = fits.ImageHDU(data = dat, header = hdr, name = 'SCI')
		cehdr = fits.ImageHDU(data = err, header = imgdat[('ERR', ext)].header, name = 'ERR')
		cdqhdr = fits.ImageHDU(data = dq, header = imgdat[('DQ', ext)].header, name = 'DQ')


	hdlist = [cphdr, cihdr, cehdr, cdqhdr]

	hdulist = fits.HDUList(hdlist)
	hdulist.writeto(img.replace('.fits', '_mask.fits'))


def cutout(img, coords, ext = 1, fov_pixel = 120, save = True):
	"""
	Get cutout of image around some coordinates.

	Parameters:
		img (str): Path to image to crop.
		coords (astropy skycoords object): Coordinates of object of interest or list of skycoord objects.
			Easiest is to feed in the output of spike.tools.objloc.
		ext (int): Integer index of extension to crop.
		fov_pixel (int): "Diameter" of square cutout region in pixels.
		save (bool): If True, will save a FITS file containing cropped region. Note that the 
			output FITS file will not have any distortion corrections stored.

	Returns: 
		cutoutim (arr): Array containing cutout region of the image.

		If save = True, will save a cropped version of the .fits file (given 
		image name + _crop suffix), including modified WCS.
	"""

	imgdat = fits.open(img)

	cphdr = fits.PrimaryHDU(header = imgdat[0].header)

	dat = imgdat[ext].data
	hdr = imgdat[ext].header

	wcs = WCS(hdr, imgdat)
	pos = utils.skycoord_to_pixel(coords, wcs)

	x0 = int(pos[0])
	y0 = int(pos[1])

	xmin = x0 - fov_pixel//2
	xmax = x0 + fov_pixel//2
	ymin = y0 - fov_pixel//2
	ymax = y0 + fov_pixel//2

	if fov_pixel % 2 == 0:
		coords0 = utils.pixel_to_skycoord(x0, y0, wcs)
		ra = coords0.ra.deg
		dec = coords0.dec.deg
		cutoutim = dat[ymin:ymax, xmin:xmax]
	if fov_pixel % 2 != 0:
		ra = coords.ra.deg
		dec = coords.dec.deg
		cutoutim = dat[ymin:ymax+1, xmin:xmax+1]

	hdr['CRVAL1'] = ra
	hdr['CRVAL2'] = dec
	hdr['CRPIX1'] = fov_pixel//2
	hdr['CRPIX2'] = fov_pixel//2
	hdr['NAXIS1'] = fov_pixel
	hdr['NAXIS2'] = fov_pixel

	# remove WCSDVARR and D2IMARR keys, since output isn't really on original
	# image grid any longer
	try:
		del hdr['DP1']
		del hdr['DP2']
		del hdr['D2IM1']
		del hdr['D2IM2']
		del hdr['D2IMDIS1']
		del hdr['D2IMDIS2']
		del hdr['D2IMERR1']
		del hdr['D2IMERR2']
		del hdr['CPDIS1']
		del hdr['CPDIS2']
		del hdr['CPERR1']
		del hdr['CPERR2']
	except:
		pass

	cihdr = fits.ImageHDU(data = cutoutim, header = hdr, name = 'SCI')

	hdlist = [cphdr, cihdr]


	hdulist = fits.HDUList(hdlist)
	if save:
		hdulist.writeto(img.replace('.fits', '_crop.fits'))

	return cutoutim




"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/resample/resample_utils.py
"""

import logging
import warnings
from typing import Tuple

import gwcs
import numpy as np
from astropy import wcs as fitswcs
from astropy.modeling import Model
from astropy.nddata.bitmask import bitfield_to_boolean_mask
from roman_datamodels.dqflags import pixel
from spike.stcal.alignment_util import wcs_from_footprints

from spike.romancal.assign_wcs_utils import wcs_bbox_from_shape

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def make_output_wcs(
    input_models,
    pscale_ratio=None,
    pscale=None,
    rotation=None,
    shape=None,
    crpix: Tuple[float, float] = None,
    crval: Tuple[float, float] = None,
):
    """
    Generate output WCS here based on footprints of all input WCS objects

    Parameters
    ----------
    input_models : list of `roman_datamodels.datamodels.DataModel`
        Each datamodel must have a `gwcs.wcs.WCS` object.

    pscale_ratio : float, optional
        Ratio of input to output pixel scale. Ignored when ``pscale`` is provided.

    pscale : float, None, optional
        Absolute pixel scale in degrees. When provided, overrides ``pscale_ratio``.

    rotation : float, None, optional
        Position angle (in degrees) of output image's Y-axis relative to North.
        A value of 0.0 would orient the final output image to be North up.
        The default of `None` specifies that the images will not be rotated,
        but will instead be resampled in the default orientation for the camera
        with the x and y axes of the resampled image corresponding
        approximately to the detector axes.

    shape : tuple of int, None, optional
        Shape of the image (data array) using ``numpy.ndarray`` convention
        (``ny`` first and ``nx`` second). This value will be assigned
        to ``pixel_shape`` and ``array_shape`` properties of the returned
        WCS object.

    crpix : tuple of float, None, optional
        Position of the reference pixel in the image array.  If ``ref_pixel`` is not
        specified, it will be set to the center of the bounding box of the
        returned WCS object.

    crval : tuple of float, None, optional
        Right ascension and declination of the reference pixel. Automatically
        computed if not provided.

    Returns
    -------
    output_wcs : object
        WCS object, with defined domain, covering entire set of input frames
    """

    wcslist = [i.meta.wcs for i in input_models]
    for w, i in zip(wcslist, input_models):
        if w.bounding_box is None:
            w.bounding_box = wcs_bbox_from_shape(i.data.shape)
    naxes = wcslist[0].output_frame.naxes

    if naxes != 2:
        raise RuntimeError(f"Output WCS needs 2 axes.{wcslist[0]} has {naxes}.")

    output_wcs = wcs_from_footprints(
        wcslist,
        None,
        dict(input_models[0].meta.wcsinfo),
        pscale_ratio=pscale_ratio,
        pscale=pscale,
        rotation=rotation,
        shape=shape,
        crpix=crpix,
        crval=crval,
    )

    # Check that the output data shape has no zero length dimensions
    if not np.prod(output_wcs.array_shape):
        raise ValueError(
            f"Invalid output frame shape: {tuple(output_wcs.array_shape)}; dimensions"
            " must have length > 0."
        )

    return output_wcs


def build_driz_weight(
    model,
    weight_type=None,
    good_bits: str = None,
):
    """
    Builds the drizzle weight map for resampling.

    Parameters
    ----------
    model : object
        The input model.
    weight_type : str, optional
        The type of weight to use. Allowed values are 'ivm' or 'exptime'.
        Defaults to None.
    good_bits : str, optional
        The good bits to use for building the mask. Defaults to None.

    Returns
    -------
    numpy.ndarray
        The drizzle weight map.

    Raises
    ------
    ValueError
        If an invalid weight type is provided.

    Examples
    --------
    .. code-block:: none

        model = get_input_model()
        weight_map = build_driz_weight(model, weight_type='ivm', good_bits=[1, 2, 3])
        print(weight_map)
    """

    dqmask = bitfield_to_boolean_mask(
        model.dq,
        good_bits,
        good_mask_value=1,
        dtype=np.uint8,
        flag_name_map=pixel,
    )

    if weight_type == "ivm":
        if (
            hasattr(model, "var_rnoise")
            and model.var_rnoise is not None
            and model.var_rnoise.shape == model.data.shape
        ):
            with np.errstate(divide="ignore", invalid="ignore"):
                inv_variance = model.var_rnoise**-1
            inv_variance[~np.isfinite(inv_variance)] = 1
        else:
            warnings.warn(
                "var_rnoise array not available. Setting drizzle weight map to 1",
                RuntimeWarning,
            )
            inv_variance = 1.0
        result = inv_variance * dqmask
    elif weight_type == "exptime":
        exptime = model.meta.exposure.exposure_time
        result = exptime * dqmask
    elif weight_type is None:
        result = np.ones(model.data.shape, dtype=model.data.dtype) * dqmask
    else:
        raise ValueError(
            f"Invalid weight type: {weight_type}."
            "Allowed weight types are 'ivm' or 'exptime'."
        )

    return result.astype(np.float32)


def build_mask(dqarr, bitvalue):
    """
    Build a bit mask from an input DQ array and a bitvalue flag.

    Parameters
    ----------
    dqarr : numpy.ndarray
        Input DQ array.
    bitvalue : str
        Bitvalue flag.

    Returns
    -------
    ndarray
        Bit mask where 1 represents good and 0 represents bad.

    Notes
    -----
    - The function interprets the bitvalue flag using the
      `astropy.nddata.bitmask.interpret_bit_flags` function.
    - If the bitvalue is None, the function returns a bit mask with all elements
      set to 1.
    - Otherwise, the function performs a bitwise AND operation between the dqarr and
      the complement of the bitvalue, and then applies a logical NOT operation to
      obtain the bit mask.
    - The resulting bit mask is returned as a `numpy.ndarray` of dtype `numpy.uint8`.
    """
    warnings.warn(
        "'build_mask' has been deprecated since release 0.16.3. "
        "Use functions from astropy.nddata.bitmask module instead such as "
        "bitfield_to_boolean_mask().",
        DeprecationWarning,
    )
    dqmask = bitfield_to_boolean_mask(
        dqarr,
        bitvalue,
        good_mask_value=1,
        dtype=np.uint8,
        flag_name_map=pixel,
    )
    return dqmask


def calc_gwcs_pixmap(in_wcs, out_wcs, shape=None):
    """
    Generate a pixel map grid using the input and output WCS.

    Parameters
    ----------
    in_wcs : `~astropy.wcs.WCS`
        Input WCS.
    out_wcs : `~astropy.wcs.WCS`
        Output WCS.
    shape : tuple, optional
        Shape of the data. If provided, the bounding box will be calculated
        from the shape. If not provided, the bounding box will be calculated
        from the input WCS.

    Returns
    -------
    pixmap : `~numpy.ndarray`
        The calculated pixel map grid.

    """
    if shape:
        bb = wcs_bbox_from_shape(shape)
        log.debug(f"Bounding box from data shape: {bb}")
    else:
        bb = in_wcs.bounding_box
        log.debug(f"Bounding box from WCS: {in_wcs.bounding_box}")

    grid = gwcs.wcstools.grid_from_bounding_box(bb)
    return np.dstack(reproject(in_wcs, out_wcs)(grid[0], grid[1]))


def reproject(wcs1, wcs2):
    """
    Given two WCSs or transforms return a function which takes pixel
    coordinates in the first WCS or transform and computes them in the second
    one. It performs the forward transformation of ``wcs1`` followed by the
    inverse of ``wcs2``.

    Parameters
    ----------
    wcs1, wcs2 : `astropy.wcs.WCS` or `gwcs.wcs.WCS` or `astropy.modeling.Model`
        WCS objects.

    Returns
    -------
    : func
        Function to compute the transformations.  It takes `(x, y)`
        positions in ``wcs1`` and returns `(x, y)` positions in ``wcs2``.
    """

    if isinstance(wcs1, fitswcs.WCS):
        forward_transform = wcs1.all_pix2world
    elif isinstance(wcs1, gwcs.WCS):
        forward_transform = wcs1.forward_transform
    elif issubclass(wcs1, Model):
        forward_transform = wcs1
    else:
        raise TypeError(
            "Expected input to be astropy.wcs.WCS or gwcs.wcs.WCS "
            "object or astropy.modeling.Model subclass"
        )

    if isinstance(wcs2, fitswcs.WCS):
        backward_transform = wcs2.all_world2pix
    elif isinstance(wcs2, gwcs.WCS):
        backward_transform = wcs2.backward_transform
    elif issubclass(wcs2, Model):
        backward_transform = wcs2.inverse
    else:
        raise TypeError(
            "Expected input to be astropy.wcs.WCS or gwcs.wcs.WCS "
            "object or astropy.modeling.Model subclass"
        )

    def _reproject(x, y):
        sky = forward_transform(x, y)
        flat_sky = []
        for axis in sky:
            flat_sky.append(axis.flatten())
        # Filter out RuntimeWarnings due to computed NaNs in the WCS
        warnings.simplefilter("ignore")
        det = backward_transform(*tuple(flat_sky))
        warnings.resetwarnings()
        det_reshaped = []
        for axis in det:
            det_reshaped.append(axis.reshape(x.shape))
        return tuple(det_reshaped)

    return _reproject


def decode_context(context, x, y):
    """
    Get 0-based indices of input images that contributed to (resampled)
    output pixel with coordinates ``x`` and ``y``.

    Parameters
    ----------
    context: numpy.ndarray
        A 3D `~numpy.ndarray` of integral data type.

    x: int, list of integers, numpy.ndarray of integers
        X-coordinate of pixels to decode (3rd index into the ``context`` array)

    y: int, list of integers, numpy.ndarray of integers
        Y-coordinate of pixels to decode (2nd index into the ``context`` array)

    Returns
    -------

    A list of `numpy.ndarray` objects each containing indices of input images
    that have contributed to an output pixel with coordinates ``x`` and ``y``.
    The length of returned list is equal to the number of input coordinate
    arrays ``x`` and ``y``.

    Examples
    --------

    An example context array for an output image of array shape ``(5, 6)``
    obtained by resampling 80 input images.

    .. code-block:: none

        con = np.array(
            [[[0, 0, 0, 0, 0, 0],
              [0, 0, 0, 36196864, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 537920000, 0, 0, 0]],
             [[0, 0, 0, 0, 0, 0,],
              [0, 0, 0, 67125536, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 163856, 0, 0, 0]],
             [[0, 0, 0, 0, 0, 0],
              [0, 0, 0, 8203, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 0, 0, 0, 0],
              [0, 0, 32865, 0, 0, 0]]],
            dtype=np.int32
        )
        decode_context(con, [3, 2], [1, 4])
        [array([ 9, 12, 14, 19, 21, 25, 37, 40, 46, 58, 64, 65, 67, 77]),
        array([ 9, 20, 29, 36, 47, 49, 64, 69, 70, 79])]

    """
    if context.ndim != 3:
        raise ValueError("'context' must be a 3D array.")

    x = np.atleast_1d(x)
    y = np.atleast_1d(y)

    if x.size != y.size:
        raise ValueError("Coordinate arrays must have equal length.")

    if x.ndim != 1:
        raise ValueError("Coordinates must be scalars or 1D arrays.")

    if not (np.issubdtype(x.dtype, np.integer) and np.issubdtype(y.dtype, np.integer)):
        raise ValueError("Pixel coordinates must be integer values")

    nbits = 8 * context.dtype.itemsize

    return [
        np.flatnonzero([v & (1 << k) for v in context[:, yi, xi] for k in range(nbits)])
        for xi, yi in zip(x, y)
    ]


def resample_range(data_shape, bbox=None):
    # Find range of input pixels to resample:
    if bbox is None:
        xmin = ymin = 0
        xmax = data_shape[1] - 1
        ymax = data_shape[0] - 1
    else:
        ((x1, x2), (y1, y2)) = bbox
        xmin = max(0, int(x1 + 0.5))
        ymin = max(0, int(y1 + 0.5))
        xmax = min(data_shape[1] - 1, int(x2 + 0.5))
        ymax = min(data_shape[0] - 1, int(y2 + 0.5))

    return xmin, xmax, ymin, ymax


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/stpipe/core.py
"""

from functools import wraps
import logging
import warnings

from stdatamodels.jwst.datamodels import JwstDataModel
from stdatamodels.jwst import datamodels
from spike.stpipe import crds_client
from spike.stpipe.step import Step
from spike.stpipe.pipeline import Pipeline

# from jwst import __version_commit__, __version__
__version_commit__ = '1.16.1' #hard coding based on released development version
__version__ = '1.16.1' # to avoid headache + keep track
from spike.jwstcal.lib_suffix import remove_suffix


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class JwstStep(Step):

    spec = """
    output_ext = string(default='.fits')  # Output file type
    """

    @classmethod
    def _datamodels_open(cls, init, **kwargs):
        return datamodels.open(init, **kwargs)


    def load_as_level2_asn(self, obj):
        """Load object as an association

        Loads the specified object into a Level2 association.
        If necessary, prepend `Step.input_dir` to all members.

        Parameters
        ----------
        obj : object
            Object to load as a Level2 association

        Returns
        -------
        association : jwst.associations.lib.rules_level2_base.DMSLevel2bBase
            Association
        """
        # Prevent circular import:
        from ..associations.load_as_asn import LoadAsLevel2Asn
        from ..associations.lib.update_path import update_key_value

        asn = LoadAsLevel2Asn.load(obj, basename=self.output_file)
        update_key_value(asn, 'expname', (), mod_func=self.make_input_path)
        return asn

    def load_as_level3_asn(self, obj):
        """Load object as an association

        Loads the specified object into a Level3 association.
        If necessary, prepend `Step.input_dir` to all members.

        Parameters
        ----------
        obj : object
            Object to load as a Level3 association

        Returns
        -------
        association : jwst.associations.lib.rules_level3_base.DMS_Level3_Base
            Association
        """
        # Prevent circular import:
        from ..associations.load_as_asn import LoadAsAssociation
        from ..associations.lib.update_path import update_key_value

        asn = LoadAsAssociation.load(obj)
        update_key_value(asn, 'expname', (), mod_func=self.make_input_path)
        return asn

    def finalize_result(self, result, reference_files_used):
        if isinstance(result, JwstDataModel):
            result.meta.calibration_software_revision = __version_commit__ or 'RELEASE'
            result.meta.calibration_software_version = __version__

            if len(reference_files_used) > 0:
                for ref_name, filename in reference_files_used:
                    if hasattr(result.meta.ref_file, ref_name):
                        getattr(result.meta.ref_file, ref_name).name = filename
                result.meta.ref_file.crds.sw_version = crds_client.get_svn_version()
                result.meta.ref_file.crds.context_used = crds_client.get_context_used(result.crds_observatory)
                if self.parent is None:
                    log.info(f"Results used CRDS context: {result.meta.ref_file.crds.context_used}")


    def remove_suffix(self, name):
        return remove_suffix(name)

    @wraps(Step.run)
    def run(self, *args, **kwargs):
        result = super().run(*args, **kwargs)
        if not self.parent:
            log.info(f"Results used jwst version: {__version__}")
        return result

    @wraps(Step.__call__)
    def __call__(self, *args, **kwargs):
        if not self.parent:
            warnings.warn(
                "Step.__call__ is deprecated. It is equivalent to Step.run "
                "and is not recommended. See "
                "https://jwst-pipeline.readthedocs.io/en/latest/jwst/"
                "user_documentation/running_pipeline_python.html"
                "#advanced-use-pipeline-run-vs-pipeline-call for more details.",
                UserWarning
            )
        return super().__call__(*args, **kwargs)


# JwstPipeline needs to inherit from Pipeline, but also
# be a subclass of JwstStep so that it will pass checks
# when constructing a pipeline using JwstStep class methods.
class JwstPipeline(Pipeline, JwstStep):
    def finalize_result(self, result, reference_files_used):
        if isinstance(result, JwstDataModel):
            log.info(f"Results used CRDS context: {crds_client.get_context_used(result.crds_observatory)}")


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/lib/dms_base.py
"""

from spike.jwstcal.associations_lib_counter import Counter

from spike.jwstcal.associations_exceptions import (
    AssociationNotValidError,
)
from spike.jwstcal.associations_lib_acid import ACIDMixin
from spike.jwstcal.associations_lib_constraint import (Constraint, AttrConstraint, SimpleConstraint)
from spike.jwstcal.associations_lib_diff import MultiDiffError, compare_asns
from spike.jwstcal.associations_lib_utilities import getattr_from_list


__all__ = ['Constraint_TargetAcq', 'Constraint_TSO', 'Constraint_WFSC', 'DMSBaseMixin']

# Default product name
PRODUCT_NAME_DEFAULT = 'undefined'

# DMS file name templates
_ASN_NAME_TEMPLATE_STAMP = 'jw{program}-{acid}_{stamp}_{type}_{sequence:05d}_asn'
_ASN_NAME_TEMPLATE = 'jw{program}-{acid}_{type}_{sequence:05d}_asn'

# Acquisition and Confirmation images
ACQ_EXP_TYPES = (
    'mir_tacq',
    'mir_taconfirm',
    'nis_taconfirm',
    'nis_tacq',
    'nrc_taconfirm',
    'nrc_tacq',
    'nrs_confirm',
    'nrs_msata',
    'nrs_taconfirm',
    'nrs_tacq',
    'nrs_taslit',
    'nrs_verify',
    'nrs_wata',
)

# Exposure EXP_TYPE to Association EXPTYPE mapping
EXPTYPE_MAP = {
    'mir_darkall':       'dark',
    'mir_darkimg':       'dark',
    'mir_darkmrs':       'dark',
    'mir_flatimage':     'flat',
    'mir_flatmrs':       'flat',
    'mir_flatimage-ext': 'flat',
    'mir_flatmrs-ext':   'flat',
    'mir_tacq':          'target_acquisition',
    'mir_taconfirm':     'target_acquisition',
    'nis_dark':          'dark',
    'nis_focus':         'engineering',
    'nis_lamp':          'engineering',
    'nis_tacq':          'target_acquisition',
    'nis_taconfirm':     'target_acquisition',
    'nrc_dark':          'dark',
    'nrc_flat':          'flat',
    'nrc_focus':         'engineering',
    'nrc_led':           'engineering',
    'nrc_tacq':          'target_acquisition',
    'nrc_taconfirm':     'target_acquisition',
    'nrs_autoflat':      'autoflat',
    'nrs_autowave':      'autowave',
    'nrs_confirm':       'target_acquisition',
    'nrs_dark':          'dark',
    'nrs_focus':         'engineering',
    'nrs_image':         'engineering',
    'nrs_lamp':          'engineering',
    'nrs_msata':         'target_acquisition',
    'nrs_tacq':          'target_acquisition',
    'nrs_taconfirm':     'target_acquisition',
    'nrs_taslit':        'target_acquisition',
    'nrs_wata':          'target_acquisition',
}

# Coronographic exposures
CORON_EXP_TYPES = [
    'mir_4qpm',
    'mir_lyot',
    'nrc_coron'
]

# Exposures that get Level2b processing
IMAGE2_SCIENCE_EXP_TYPES = [
    'fgs_image',
    'mir_4qpm',
    'mir_image',
    'mir_lyot',
    'nis_ami',
    'nis_image',
    'nrc_coron',
    'nrc_image',
    'nrs_mimf',
    'nrc_tsimage',
]

IMAGE2_NONSCIENCE_EXP_TYPES = [
    'mir_coroncal',
    'nis_focus',
    'nrc_focus',
    'nrs_focus',
    'nrs_image',
]
IMAGE2_NONSCIENCE_EXP_TYPES.extend(ACQ_EXP_TYPES)

SPEC2_SCIENCE_EXP_TYPES = [
    'mir_lrs-fixedslit',
    'mir_lrs-slitless',
    'mir_mrs',
    'nis_soss',
    'nis_wfss',
    'nrc_tsgrism',
    'nrc_wfss',
    'nrs_fixedslit',
    'nrs_ifu',
    'nrs_msaspec',
    'nrs_brightobj',
]

# Modifiers from the pool that define the primary use of
# an exposure.
#
# Note: Order is important with the first items taking
# higher precedence. This depends on Python dicts ordering
# by order of key addition.
SPECIAL_EXPOSURE_MODIFIERS = {
    'psf': ['is_psf'],
    'imprint': ['is_imprt'],
    'background': ['bkgdtarg'],
}

# Exposures that are always TSO
TSO_EXP_TYPES = [
    'nrc_tsimage',
    'nrc_tsgrism',
    'nrs_brightobj'
]

# Define the valid optical paths vs detector for NIRSpec Fixed-slit Science
# Tuples are (SLIT, GRATING, FILTER, DETECTOR)
# All A-slits are represented by SLIT == 'a'.
NRS_FSS_VALID_OPTICAL_PATHS = (
    ('a', 'prism', 'clear',  'nrs1'),
    ('a', 'g395h', 'f290lp', 'nrs1'),
    ('a', 'g395h', 'f290lp', 'nrs2'),
    ('a', 'g235h', 'f170lp', 'nrs1'),
    ('a', 'g235h', 'f170lp', 'nrs2'),
    ('a', 'g140h', 'f100lp', 'nrs1'),
    ('a', 'g140h', 'f100lp', 'nrs2'),
    ('a', 'g140h', 'f070lp', 'nrs1'),
    ('a', 'g395m', 'f290lp', 'nrs1'),
    ('a', 'g235m', 'f170lp', 'nrs1'),
    ('a', 'g140m', 'f100lp', 'nrs1'),
    ('a', 'g140m', 'f070lp', 'nrs1'),
    ('s200b1', 'prism', 'clear',  'nrs2'),
    ('s200b1', 'g395h', 'f290lp', 'nrs2'),
    ('s200b1', 'g235h', 'f170lp', 'nrs2'),
    ('s200b1', 'g140h', 'f100lp', 'nrs2'),
    ('s200b1', 'g140h', 'f070lp', 'nrs1'),
    ('s200b1', 'g140h', 'f070lp', 'nrs2'),
    ('s200b1', 'g395m', 'f290lp', 'nrs2'),
    ('s200b1', 'g235m', 'f170lp', 'nrs2'),
    ('s200b1', 'g140m', 'f100lp', 'nrs2'),
    ('s200b1', 'g140m', 'f070lp', 'nrs1'),
    ('s200b1', 'g140m', 'f070lp', 'nrs2'),
)

NRS_FSS_VALID_LAMP_OPTICAL_PATHS = (
    ('prism', 'line4', 'nrs1'),
    ('prism', 'line4', 'nrs2'),
    ('prism', 'flat5', 'nrs1'),
    ('prism', 'flat5', 'nrs2'),
    ('prism', 'test', 'nrs1'),
    ('prism', 'test', 'nrs2'),
    ('g395h', 'flat3', 'nrs1'),
    ('g395h', 'flat3', 'nrs2'),
    ('g395h', 'line3', 'nrs1'),
    ('g395h', 'line3', 'nrs2'),
    ('g235h', 'flat2', 'nrs1'),
    ('g235h', 'flat2', 'nrs2'),
    ('g235h', 'line2', 'nrs1'),
    ('g235h', 'line2', 'nrs2'),
    ('g140h', 'flat1', 'nrs1'),
    ('g140h', 'flat1', 'nrs2'),
    ('g140h', 'ref', 'nrs1'),
    ('g140h', 'ref', 'nrs2'),
    ('g140h', 'line1', 'nrs1'),
    ('g140h', 'line1', 'nrs2'),
    ('g140h', 'flat4', 'nrs1'),
    ('g140h', 'flat4', 'nrs2'),
    ('g395m', 'flat3', 'nrs1'),
    ('g395m', 'flat3', 'nrs2'),
    ('g395m', 'line3', 'nrs1'),
    ('g395m', 'line3', 'nrs2'),
    ('g235m', 'flat2', 'nrs1'),
    ('g235m', 'flat2', 'nrs2'),
    ('g235m', 'line2', 'nrs1'),
    ('g235m', 'line2', 'nrs2'),
    ('g140m', 'flat1', 'nrs1'),
    ('g140m', 'flat1', 'nrs2'),
    ('g140m', 'line1', 'nrs1'),
    ('g140m', 'line1', 'nrs2'),
    ('g140m', 'flat4', 'nrs1'),
    ('g140m', 'flat4', 'nrs2'),
)

# Define the valid optical paths vs detector for NIRSpec IFU Science
# Tuples are (GRATING, FILTER, DETECTOR)
# The only combinations that result in data on the NRS2 detector are
# g140h/f100lp, g235h/f170lp, and g395h/f290lp.
NRS_IFU_VALID_OPTICAL_PATHS = (
    ('prism', 'clear',  'nrs1'),
    ('g140m', 'f070lp', 'nrs1'),
    ('g140m', 'f100lp', 'nrs1'),
    ('g235m', 'f170lp', 'nrs1'),
    ('g395m', 'f290lp', 'nrs1'),
    ('g140h', 'f070lp', 'nrs1'),
    ('g140h', 'f100lp', 'nrs1'),
    ('g140h', 'f100lp', 'nrs2'),
    ('g235h', 'f170lp', 'nrs1'),
    ('g235h', 'f170lp', 'nrs2'),
    ('g395h', 'f290lp', 'nrs1'),
    ('g395h', 'f290lp', 'nrs2'),
)


# Define the uncalibrated filters for NIRISS SOSS
# Data taken with these filters should not be processed beyond
# level 2a.
NIS_SOSS_UNCALIBRATED_FILTERS = ('f277w',)


# Key that uniquely identifies members.
MEMBER_KEY = 'expname'

# Non-specified values found in DMS Association Pools
_EMPTY = (None, '', 'NULL', 'Null', 'null', '--', 'N', 'n',
          'F', 'f', 'FALSE', 'false', 'False', 'N/A', 'n/a')

# Degraded status information
_DEGRADED_STATUS_OK = (
    'No known degraded exposures in association.'
)
_DEGRADED_STATUS_NOTOK = (
    'One or more members have an error associated with them.'
    '\nDetails can be found in the member.exposerr attribute.'
)


class DMSBaseMixin(ACIDMixin):
    """Association attributes common to DMS-based Rules

    Attributes
    ----------
    sequence : int
        The sequence number of the current association
    """

    # Associations of the same type are sequenced.
    _sequence = Counter(start=1)

    def __init__(self, *args, **kwargs):
        super(DMSBaseMixin, self).__init__(*args, **kwargs)

        self._acid = None
        self._asn_name = None
        self.sequence = None
        if 'degraded_status' not in self.data:
            self.data['degraded_status'] = _DEGRADED_STATUS_OK
        if 'program' not in self.data:
            self.data['program'] = 'noprogram'

    @classmethod
    def create(cls, item, version_id=None):
        """Create association if item belongs

        Parameters
        ----------
        item : dict
            The item to initialize the association with.

        version_id : str or None
            Version_Id to use in the name of this association.
            If None, nothing is added.

        Returns
        -------
        (association, reprocess_list)
            2-tuple consisting of:

                - association : The association or, if the item does not
                  match this rule, None
                - [ProcessList[, ...]]: List of items to process again.
        """
        asn, reprocess = super(DMSBaseMixin, cls).create(item, version_id)
        if not asn:
            return None, reprocess
        asn.sequence = next(asn._sequence)
        return asn, reprocess

    @property
    def acid(self):
        """Association ID"""
        acid = self._acid
        if self._acid is None:
            acid = self.acid_from_constraints()
        return acid

    @property
    def asn_name(self):
        """The association name

        The name that identifies this association. When dumped,
        will form the basis for the suggested file name.

        Typically, it is generated based on the current state of
        the association, but can be overridden.
        """
        if self._asn_name:
            return self._asn_name

        program = self.data['program']
        version_id = self.version_id
        asn_type = self.data['asn_type']
        sequence = self.sequence

        if version_id:
            name = _ASN_NAME_TEMPLATE_STAMP.format(
                program=program,
                acid=self.acid.id,
                stamp=version_id,
                type=asn_type,
                sequence=sequence,
            )
        else:
            name = _ASN_NAME_TEMPLATE.format(
                program=program,
                acid=self.acid.id,
                type=asn_type,
                sequence=sequence,
            )
        return name.lower()

    @asn_name.setter
    def asn_name(self, name):
        """Override calculated association name"""
        self._asn_name = name

    @property
    def current_product(self):
        return self.data['products'][-1]

    @property
    def from_items(self):
        """The list of items that contributed to the association."""
        try:
            items = [
                member.item
                for product in self['products']
                for member in product['members']
            ]
        except KeyError:
            items = []
        return items

    @property
    def member_ids(self):
        """Set of all member ids in all products of this association"""
        member_ids = set(
            member[MEMBER_KEY]
            for product in self['products']
            for member in product['members']
        )
        return member_ids

    @property
    def validity(self):
        """Keeper of the validity tests"""
        try:
            validity = self._validity
        except AttributeError:
            self._validity = {}
            validity = self._validity
        return validity

    @validity.setter
    def validity(self, item):
        """Set validity dict"""
        self._validity = item

    def get_exposure_type(self, item, default='science'):
        """Determine the exposure type of a pool item

        Parameters
        ----------
        item : dict
            The pool entry to determine the exposure type of

        default : str or None
            The default exposure type.
            If None, routine will raise LookupError

        Returns
        -------
        exposure_type : str
            Exposure type. Can be one of

                - 'science': Item contains science data
                - 'target_acquisition': Item contains target acquisition data.
                - 'autoflat': NIRSpec AUTOFLAT
                - 'autowave': NIRSpec AUTOWAVE
                - 'psf': PSF
                - 'imprint': MSA/IFU Imprint/Leakcal

        Raises
        ------
        LookupError
            When `default` is None and an exposure type cannot be determined
        """
        return get_exposure_type(item, default=default, association=self)

    def is_member(self, new_member):
        """Check if member is already a member

        Parameters
        ----------
        new_member : Member
            The member to check for
        """
        try:
            current_members = self.current_product['members']
        except KeyError:
            return False

        for member in current_members:
            if member == new_member:
                return True
        return False

    def is_item_member(self, item):
        """Check if item is already a member of this association

        Parameters
        ----------
        item : dict
            The item to check for.

        Returns
        -------
        is_item_member : bool
            True if item is a member.
        """
        return item in self.from_items

    def is_item_ami(self, item):
        """Is the given item AMI (NIRISS Aperture Masking Interferometry)

        Determine whether the specific item represents AMI data or not.
        This simply includes items with EXP_TYPE='NIS_AMI'.

        Parameters
        ----------
        item : dict
            The item to check for.

        Returns
        -------
        is_item_ami : bool
            Item represents an AMI exposure.
        """
        # If not a science exposure, such as target acquisitions,
        # then other indicators do not apply.
        if item['pntgtype'] != 'science':
            return False

        # Target acquisitions are never AMI
        if item['exp_type'] in ACQ_EXP_TYPES:
            return False

        # Check for AMI exposure type
        try:
            is_ami = self.item_getattr(item, ['exp_type'])[1] in ['nis_ami']
        except KeyError:
            is_ami = False

        return is_ami

    def is_item_coron(self, item):
        """Is the given item Coronagraphic

        Determine whether the specific item represents
        true Coronagraphic data or not. This will include all items
        in CORON_EXP_TYPES (both NIRCam and MIRI), **except** for
        NIRCam short-wave detectors included in a coronagraphic exposure
        but do not have an occulter in their field-of-view.

        Parameters
        ----------
        item : dict
            The item to check for.

        Returns
        -------
        is_item_coron : bool
            Item represents a true Coron exposure.
        """
        # If not a science exposure, such as target acquisitions,
        # then other indicators do not apply.
        if item['pntgtype'] != 'science':
            return False

        # Target acquisitions are never Coron
        if item['exp_type'] in ACQ_EXP_TYPES:
            return False

        # Check for coronagraphic exposure type
        try:
            is_coron = self.item_getattr(item, ['exp_type'])[1] in CORON_EXP_TYPES
        except KeyError:
            is_coron = False
            return is_coron

        # Now do a special check for NRC_CORON exposures using full-frame readout,
        # which include extra detectors that do *not* have an occulter in them
        if item['exp_type'] == 'nrc_coron' and item['subarray'] == 'full':
            if item['pupil'] == 'maskbar' and item['detector'] in ['nrca1', 'nrca2', 'nrca3']:
                is_coron = False
            if item['pupil'] == 'maskrnd' and item['detector'] in ['nrca1', 'nrca3', 'nrca4']:
                is_coron = False

        return is_coron

    def is_item_tso(self, item, other_exp_types=None):
        """Is the given item TSO

        Determine whether the specific item represents
        TSO data or not. This is used to determine the
        naming of files, i.e. "rate" vs "rateints" and
        "cal" vs "calints".

        Parameters
        ----------
        item : dict
            The item to check for.

        other_exp_types: [str[,...]] or None
            List of other exposure types to consider TSO-like.

        Returns
        -------
        is_item_tso : bool
            Item represents a TSO exposure.
        """
        # If not a science exposure, such as target acquisitions,
        # then other TSO indicators do not apply.
        if item['pntgtype'] != 'science':
            return False

        # Target acquisitions are never TSO
        if item['exp_type'] in ACQ_EXP_TYPES:
            return False

        # Setup exposure list
        all_exp_types = TSO_EXP_TYPES.copy()
        if other_exp_types:
            all_exp_types += other_exp_types

        # Go through all other TSO indicators.
        try:
            is_tso = self.constraints['is_tso'].value == 't'
        except (AttributeError, KeyError):
            # No such constraint is defined. Just continue on.
            is_tso = False
        try:
            is_tso = is_tso or self.item_getattr(item, ['tsovisit'])[1] == 't'
        except KeyError:
            pass
        try:
            is_tso = is_tso or self.item_getattr(item, ['exp_type'])[1] in all_exp_types
        except KeyError:
            pass
        return is_tso

    def item_getattr(self, item, attributes):
        """Return value from any of a list of attributes

        Parameters
        ----------
        item : dict
            item to retrieve from

        attributes : list
            List of attributes

        Returns
        -------
        (attribute, value)
            Returns the value and the attribute from
            which the value was taken.

        Raises
        ------
        KeyError
            None of the attributes are found in the dict.
        """
        return item_getattr(item, attributes, self)

    def new_product(self, product_name=PRODUCT_NAME_DEFAULT):
        """Start a new product"""
        product = {
            'name': product_name,
            'members': []
        }
        try:
            self.data['products'].append(product)
        except (AttributeError, KeyError):
            self.data['products'] = [product]

    def update_asn(self, item=None, member=None):
        """Update association meta information

        Parameters
        ----------
        item : dict or None
            Item to use as a source. If not given, item-specific
            information will be left unchanged.

        member : Member or None
            An association member to use as source.
            If not given, member-specific information will be update
            from current association/product membership.

        Notes
        -----
        If both `item` and `member` are given,
        information in `member` will take precedence.
        """
        self.update_degraded_status()

    def update_degraded_status(self):
        """Update association degraded status"""

        if self.data['degraded_status'] == _DEGRADED_STATUS_OK:
            for product in self.data['products']:
                for member in product['members']:
                    try:
                        exposerr = member['exposerr']
                    except KeyError:
                        continue
                    else:
                        if exposerr not in _EMPTY:
                            self.data['degraded_status'] = _DEGRADED_STATUS_NOTOK
                            break

    def update_validity(self, entry):
        for test in self.validity.values():
            if not test['validated']:
                test['validated'] = test['check'](entry)

    @classmethod
    def reset_sequence(cls):
        cls._sequence = Counter(start=1)

    @classmethod
    def validate(cls, asn):
        super(DMSBaseMixin, cls).validate(asn)

        if isinstance(asn, DMSBaseMixin):
            result = False
            try:
                result = all(
                    test['validated']
                    for test in asn.validity.values()
                )
            except (AttributeError, KeyError):
                raise AssociationNotValidError('Validation failed')
            if not result:
                raise AssociationNotValidError(
                    'Validation failed validity tests.'
                )

        return True

    def _get_exposure(self):
        """Get string representation of the exposure id

        Returns
        -------
        exposure : str
            The Level3 Product name representation
            of the exposure & activity id.
        """
        exposure = ''
        try:
            activity_id = format_list(
                self.constraints['activity_id'].found_values
            )
        except KeyError:
            pass
        else:
            if activity_id not in _EMPTY:
                exposure = '{0:0>2s}'.format(activity_id)
        return exposure

    def _get_instrument(self):
        """Get string representation of the instrument

        Returns
        -------
        instrument : str
            The Level3 Product name representation
            of the instrument
        """
        instrument = format_list(self.constraints['instrument'].found_values)
        return instrument

    def _get_opt_element(self):
        """Get string representation of the optical elements.
        This includes only elements contained in the filter/pupil
        wheels of the instrument.

        Returns
        -------
        opt_elem : str
            The Level3 Product name representation
            of the optical elements.
        """
        # Retrieve all the optical elements
        opt_elems = []
        for opt_elem in ['opt_elem', 'opt_elem2']:
            try:
                values = list(self.constraints[opt_elem].found_values)
            except KeyError:
                pass
            else:
                values.sort(key=str.lower)
                value = format_list(values)
                if value not in _EMPTY:
                    opt_elems.append(value)

        # Build the string. Sort the elements in order to
        # create data-independent results
        opt_elems.sort(key=str.lower)
        opt_elem = '-'.join(opt_elems)
        if opt_elem == '':
            opt_elem = 'clear'

        return opt_elem

    def _get_slit_name(self):
        """Get string representation of the slit name (NIRSpec fixed-slit only)

        Returns
        -------
        slit_name : str
            The Level3 Product name representation
            of the slit name.
        """
        # Retrieve all the slit names (sometimes there can be 2)
        slit_names = []
        for fxd_slit in ['fxd_slit', 'fxd_slit2']:
            try:
                values = list(self.constraints[fxd_slit].found_values)
            except KeyError:
                pass
            else:
                values.sort(key=str.lower)
                value = format_list(values)
                if value not in _EMPTY and value not in slit_names:
                    slit_names.append(value)

        # Build the string. Sort the elements in order to
        # create data-independent results
        slit_names.sort(key=str.lower)
        slit_name = '-'.join(slit_names)

        # Slit name may be empty string
        return slit_name

    def _get_subarray(self):
        """Get string representation of the subarray

        Returns
        -------
        subarray : str
            The Level3 Product name representation
            of the subarray.
        """
        result = ''
        try:
            subarray = format_list(self.constraints['subarray'].found_values)
        except KeyError:
            subarray = None
        if subarray == 'full':
            subarray = None
        if subarray is not None:
            result = subarray

        return result

    def _get_target(self):
        """Get string representation of the target

        Returns
        -------
        target : str
            The Level3 Product name representation
            of the target or source ID.
        """
        attrs = self.constraints.get_all_attr('found_values', name='target')
        if attrs:
            value = attrs[0][1]
        else:
            value = []
        target_id = format_list(value)
        target = 't{0:0>3s}'.format(str(target_id))
        return target

    def _get_grating(self):
        """Get string representation of the grating in use

        Returns
        -------
        grating : str
            The Level3 Product name representation
            of the grating in use.
        """
        grating_id = format_list(self.constraints['grating'].found_values)
        grating = '{0:0>3s}'.format(str(grating_id))
        return grating

    def __eq__(self, other):
        """Compare equality of two associations"""
        result = NotImplemented
        if isinstance(other, DMSBaseMixin):
            try:
                compare_asns(self, other)
            except MultiDiffError:
                result = False
            else:
                result = True

        return result

    def __ne__(self, other):
        """Compare inequality of two associations"""
        if isinstance(other, DMSBaseMixin):
            return not self.__eq__(other)

        return NotImplemented


# -----------------
# Basic constraints
# -----------------
class DMSAttrConstraint(AttrConstraint):
    """DMS-focused attribute constraint

    Forces definition of invalid values
    """
    def __init__(self, **kwargs):

        if kwargs.get('invalid_values', None) is None:
            kwargs['invalid_values'] = _EMPTY

        super(DMSAttrConstraint, self).__init__(**kwargs)


class Constraint_TargetAcq(SimpleConstraint):
    """Select on target acquisition exposures

    Parameters
    ----------
    association:  ~jwst.associations.Association
        If specified, use the `get_exposure_type` method
        of the association rather than the utility version.
    """
    def __init__(self, association=None):
        if association is None:
            _get_exposure_type = get_exposure_type
        else:
            _get_exposure_type = association.get_exposure_type

        super(Constraint_TargetAcq, self).__init__(
            name='target_acq',
            value='target_acquisition',
            sources=_get_exposure_type
        )


class Constraint_TSO(Constraint):
    """Match on Time-Series Observations"""
    def __init__(self, *args, **kwargs):
        super(Constraint_TSO, self).__init__(
            [
                DMSAttrConstraint(
                    sources=['pntgtype'],
                    value='science'
                ),
                Constraint(
                    [
                        DMSAttrConstraint(
                            sources=['tsovisit'],
                            value='t',
                        ),
                        DMSAttrConstraint(
                            sources=['exp_type'],
                            value='|'.join(TSO_EXP_TYPES),
                        ),
                    ],
                    reduce=Constraint.any
                )
            ],
            name='is_tso'
        )


class Constraint_WFSC(Constraint):
    """Match on Wave Front Sensing and Control Observations"""
    def __init__(self, *args, **kwargs):
        super(Constraint_WFSC, self).__init__(
            [
                Constraint(
                    [
                        DMSAttrConstraint(
                            name='wfsc',
                            sources=['visitype'],
                            value='.+wfsc.+',
                            force_unique=True
                        )
                    ]
                )
            ]
        )


# #########
# Utilities
# #########
def format_list(alist):
    """Format a list according to DMS naming specs"""
    return '-'.join(alist)


def get_exposure_type(item, default='science', association=None):
    """Determine the exposure type of a pool item

    Parameters
    ----------
    item : dict
        The pool entry to determine the exposure type of

    default : str or None
        The default exposure type.
        If None, routine will raise LookupError



    Returns
    -------
    exposure_type : str
        Exposure type. Can be one of

        - 'science': Item contains science data
        - 'target_acquisition': Item contains target acquisition data.
        - 'autoflat': NIRSpec AUTOFLAT
        - 'autowave': NIRSpec AUTOWAVE
        - 'psf': PSF
        - 'imprint': MSA/IFU Imprint/Leakcal

    Raises
    ------
    LookupError
        When `default` is None and an exposure type cannot be determined
    """
    # Specify how attributes of the item are retrieved.
    def _item_attr(item, sources):
        """Get attribute value of an item

        This simplifies the call to `item_getattr`
        """
        source, value = item_getattr(item, sources, association=association)
        return value

    # Define default type.
    result = default

    # Retrieve pointing type. This decides the basic exposure type.
    # If the pointing is not science, we're done.
    try:
        result = _item_attr(item, ['pntgtype'])
    except KeyError:
        pass
    else:
        if result != 'science':
            return result

    # We have a science exposure. Refine further.
    #
    # Base type off of exposure type.
    try:
        exp_type = _item_attr(item, ['exp_type'])
    except KeyError:
        raise LookupError('Exposure type cannot be determined')

    result = EXPTYPE_MAP.get(exp_type, default)

    if result is None:
        raise LookupError('Cannot determine exposure type')

    # If result is not science, we're done.
    if result != 'science':
        return result

    # For `science` data, compare against special modifiers
    # to further refine the type.
    for special, source in SPECIAL_EXPOSURE_MODIFIERS.items():
        try:
            _item_attr(item, source)
        except KeyError:
            pass
        else:
            result = special
            break

    return result


def item_getattr(item, attributes, association=None):
    """Return value from any of a list of attributes

    Parameters
    ----------
    item : dict
        item to retrieve from

    attributes : list
        List of attributes

    Returns
    -------
    (attribute, value)
        Returns the value and the attribute from
        which the value was taken.

    Raises
    ------
    KeyError
        None of the attributes are found in the dict.
    """
    if association is None:
        invalid_values = _EMPTY
    else:
        invalid_values = association.INVALID_VALUES
    return getattr_from_list(
        item,
        attributes,
        invalid_values=invalid_values
    )


def nrsfss_valid_detector(item):
    """Check that a slit/grating/filter combo can appear on the detector"""
    try:
        _, detector = item_getattr(item, ['detector'])
        _, filter = item_getattr(item, ['filter'])
        _, grating = item_getattr(item, ['grating'])
        _, slit = item_getattr(item, ['fxd_slit'])
    except KeyError:
        return False

    # Reduce all A slits to just 'a'.
    if slit != 's200b1':
        slit = 'a'

    return (slit, grating, filter, detector) in NRS_FSS_VALID_OPTICAL_PATHS


def nrsifu_valid_detector(item):
    """Check that a grating/filter combo can appear on the detector"""
    _, exp_type = item_getattr(item, ['exp_type'])
    if exp_type != 'nrs_ifu':
        return True

    try:
        _, detector = item_getattr(item, ['detector'])
        _, filter = item_getattr(item, ['filter'])
        _, grating = item_getattr(item, ['grating'])
    except KeyError:
        return False

    return (grating, filter, detector) in NRS_IFU_VALID_OPTICAL_PATHS


def nrslamp_valid_detector(item):
    """Check that a grating/lamp combo can appear on the detector"""
    try:
        _, detector = item_getattr(item, ['detector'])
        _, grating = item_getattr(item, ['grating'])
        _, opmode = item_getattr(item, ['opmode'])
    except KeyError:
        return False

    if opmode in ['msaspec']:
        # All settings can result in data on both detectors,
        # depending on which MSA shutters are open
        return True

    elif opmode in ['ifu']:
        # Need lamp value for IFU mode
        try:
            _, lamp = item_getattr(item, ['lamp'])
        except KeyError:
            return False

        # Just a checklist of paths:
        if grating in ['g395h', 'g235h']:
            # long-wave, high-res gratings result in data on both detectors
            return True
        elif grating in ['g395m', 'g235m', 'g140m'] and detector == 'nrs1':
            # all medium-res gratings result in data only on NRS1 detector
            return True
        elif grating == 'prism' and detector == 'nrs1':
            # prism results in data only on NRS1 detector
            return True
        elif grating == 'g140h':
            # short-wave, high-res grating results in data on both detectors,
            # except when lamp FLAT4 is in use (no data on NRS2)
            if not (detector == 'nrs2' and lamp == 'flat4'):
                return True

    elif opmode in ['fixedslit']:
        # All slits illuminated by lamps, regardless of grating or subarray
        try:
            _, lamp = item_getattr(item, ['lamp'])
        except KeyError:
            return False

        return (grating, lamp, detector) in NRS_FSS_VALID_LAMP_OPTICAL_PATHS

    # Nothing has matched. Not valid.
    return False


def nrccoron_valid_detector(item):
    """Check that a coronagraphic mask+detector combo is valid"""
    try:
        _, detector = item_getattr(item, ['detector'])
        _, subarray = item_getattr(item, ['subarray'])
        _, pupil = item_getattr(item, ['pupil'])
    except KeyError:
        return False

    # Just a checklist of paths:
    if subarray == 'full':
        # maskbar has occulted target only in detector nrca4
        if pupil == 'maskbar' and detector in ['nrca1', 'nrca2', 'nrca3']:
            return False
        # maskrnd has occulted target only in detector nrca2
        elif pupil == 'maskrnd' and detector in ['nrca1', 'nrca3', 'nrca4']:
            return False
        else:
            return True
    else:
        return True


def nissoss_calibrated_filter(item):
    """Check that a NIRISS filter is calibrated."""
    _, exp_type = item_getattr(item, ['exp_type'])
    if exp_type != 'nis_soss':
        return True

    try:
        _, filter = item_getattr(item, ['filter'])
    except KeyError:
        return False

    return filter not in NIS_SOSS_UNCALIBRATED_FILTERS


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/exceptions.py
"""


__all__ = [
    'AssociationError',
    'AssociationNotAConstraint',
    'AssociationNotValidError',
]


class AssociationError(Exception):
    """Basic errors related to Associations"""


class AssociationNotAConstraint(AssociationError):
    """No matching constraint found"""


class AssociationNotValidError(AssociationError):
    """Given data structure is not a valid association"""